#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2016 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import collections
from .FieldElement import FieldElement
from .AffineCurvePoint import AffineCurvePoint
from .EllipticCurve import EllipticCurve
from .DocInherit import doc_inherit
from .CurveOps import CurveOpIsomorphism, CurveOpExportSage

_ShortWeierstrassCurveDomainParameters = collections.namedtuple("ShortWeierstrassCurveDomainParameters", [ "curvetype", "a", "b", "p", "n", "h", "G" ])

class ShortWeierstrassCurve(EllipticCurve, CurveOpIsomorphism, CurveOpExportSage):
	"""Represents an elliptic curve over a finite field F_P that satisfies the
	short Weierstrass equation y^2 = x^3 + ax + b."""
	pretty_name = "Short Weierstrass"

	def __init__(self, a, b, p, n, h, Gx, Gy, **kwargs):
		"""Create an elliptic curve given the equation coefficients a and b,
		the curve modulus p, the order of the curve n, the cofactor of the
		curve h and the generator point G's X and Y coordinates in affine
		representation, Gx and Gy."""
		EllipticCurve.__init__(self, p, n, h, Gx, Gy, **kwargs)
		assert(isinstance(a, int))		# Curve coefficent A
		assert(isinstance(b, int))		# Curve coefficent B
		self._a = FieldElement(a, p)
		self._b = FieldElement(b, p)
		self._name = kwargs.get("name")

		# Check that the curve is not singular
		assert((4 * (self.a ** 3)) + (27 * (self.b ** 2)) != 0)

		if self._G is not None:
			# Check that the generator G is on the curve
			assert(self._G.oncurve())

			if self.n is not None:
				# Check that the generator G is of curve order if a order was
				# passed as well
				assert((self.n * self.G).is_neutral)

	@classmethod
	def init_rawcurve(cls, a, b, p):
		"""Returns a raw curve which has an undiscovered amount of points
		#E(F_p) (i.e. the domain parameters n and h are not set). This function
		can be used to create a curve which is later completed by counting
		#E(F_p) using Schoof's algorithm."""
		return cls(a = a, b = b, p = p, n = None, h = None, Gx = None, Gy = None)

	@property
	def is_anomalous(self):
		"""Returns if the curve is anomalous, i.e. if #F(p) == p. If this is
		the case then there is an efficient method to solve the ECDLP.
		Therefore the curve is not suitable for cryptographic use."""
		return self.jinv in [ 0, 1728 ]

	@property
	@doc_inherit(EllipticCurve)
	def domainparams(self):
		return _ShortWeierstrassCurveDomainParameters(curvetype = self.curvetype, a = self.a, b = self.b, p = self.p, n = self.n, h = self.h, G = self.G)

	@property
	@doc_inherit(EllipticCurve)
	def curvetype(self):
		return "shortweierstrass"

	@property
	def is_koblitz(self):
		"""Returns whether the curve allows for efficient computation of a map
		\phi in the field (i.e. that the curve is commonly known as a 'Koblitz
		Curve'). This corresponds to examples 3 and 4 of the paper "Faster
		Point Multiplication on Elliptic Curves with Efficient Endomorphisms"
		by Gallant, Lambert and Vanstone."""
		return ((self.b == 0) and ((self.p % 4) == 1)) or ((self.a == 0) and ((self.p % 3) == 1))

	@property
	def security_bit_estimate(self):
		"""Returns the bit security estimate of the curve. Subtracts four bits
		security margin for Koblitz curves."""
		security_bits = self.n.bit_length() // 2
		if self.is_koblitz:
			security_bits -= 4
		return security_bits

	@property
	@doc_inherit(EllipticCurve)
	def prettyname(self):
		name = [ ]
		name.append(self.pretty_name)
		if self.is_koblitz:
			name.append("(Koblitz)")
		return " ".join(name)
	
	@property
	def a(self):
		"""Returns the coefficient a of the curve equation y^2 = x^3 + ax + b."""
		return self._a

	@property
	def b(self):
		"""Returns the coefficient b of the curve equation y^2 = x^3 + ax + b."""
		return self._b

	@property
	def jinv(self):
		"""Returns the j-invariant of the curve, i.e. 1728 * 4 * a^3 / (4 * a^3
		+ 27 * b^2)."""
		return 1728 * (4 * self.a ** 3) // ((4 * self.a ** 3) + (27 * self.b ** 2))

	def getpointwithx(self, x):
		"""Returns a tuple of two points which fulfill the curve equation or
		None if not such points exist."""
		assert(isinstance(x, int))
		yy = ((FieldElement(x, self._p) ** 3) + (self._a * x) + self._b)
		y = yy.sqrt()
		if y:
			return (AffineCurvePoint(x, int(y[0]), self), AffineCurvePoint(x, int(y[1]), self))
		else:
			return None

	@doc_inherit(EllipticCurve)
	def oncurve(self, P):
		return P.is_neutral or ((P.y ** 2) == (P.x ** 3) + (self.a * P.x) + self.b)

	@doc_inherit(EllipticCurve)
	def point_conjugate(self, P):
		return AffineCurvePoint(int(P.x), int(-P.y), self)

	@doc_inherit(EllipticCurve)
	def point_addition(self, P, Q):
		if P.is_neutral:
			# P is at infinity, O + Q = Q
			result = Q
		elif Q.is_neutral:
			# Q is at infinity, P + O = P
			result = P
		elif P == -Q:
			# P == -Q, return O (point at infinity)
			result = self.neutral()
		elif P == Q:
			# P == Q, point doubling
			s = ((3 * P.x ** 2) + self.a) // (2 * P.y)
			newx = s * s - (2 * P.x)
			newy = s * (P.x - newx) - P.y
			result = AffineCurvePoint(int(newx), int(newy), self)
		else:
			# P != Q, point addition
			s = (P.y - Q.y) // (P.x - Q.x)
			newx = (s ** 2) - P.x - Q.x
			newy = s * (P.x - newx) - P.y
			result = AffineCurvePoint(int(newx), int(newy), self)
		return result

	@doc_inherit(EllipticCurve)
	def compress(self, P):
		return (int(P.x), int(P.y) % 2)

	@doc_inherit(EllipticCurve)
	def uncompress(self, compressed):
		(x, ybit) = compressed
		x = FieldElement(x, self.p)
		alpha = (x ** 3) + (self.a * x) + self.b
		(beta1, beta2) = alpha.sqrt()
		if (int(beta1) % 2) == ybit:
			y = beta1
		else:
			y = beta2
		return AffineCurvePoint(int(x), int(y), self)

	@doc_inherit(EllipticCurve)
	def enumerate_points(self):
		yield self.neutral()
		for x in range(self.p):
			points = self.getpointwithx(x)
			if points is not None:
				yield points[0]
				yield points[1]

	def __str__(self):
		if self.hasname:
			return "ShortWeierstrassCurve<%s>" % (self.name)
		else:
			return "ShortWeierstrassCurve<y^2 = x^3 + 0x%x x + 0x%x mod 0x%x>" % (int(self.a), int(self.b), int(self.p))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

from .FieldElement import FieldElement
from .Exceptions import NoSuchCurveException

class CurveOpIsomorphism(object):
	def _twist(self, d = None, sqrt_d = None):
		"""Returns the twisted curve with the twist coefficient d. If d is a
		quadratic non-residue mod p then this function will yield a curve that
		is isomorphous on the field extension GF(sqrt(d)). If it is a quadratic
		residue, it will return an GF(p)-isomorphous curve."""
		assert(self.curvetype == "shortweierstrass")
		ShortWeierstrassCurve = self.__class__

		if sqrt_d is not None:
			# If a square root is given, then it must be a correct square root
			assert(sqrt_d ** 2 == d)

		a = self.a * (d ** 2)
		b = self.b * (d ** 3)
		if d.is_qr and self.hasgenerator:
			# Quadratic twist will return an GF(p)-isomorphous curve -> convert
			# generator point as well
			if sqrt_d is None:
				sqrt_d = d.sqrt()[0]
			Gx = int(self.G.x * d)
			Gy = int(self.G.y * (sqrt_d ** 3))

			n = self.n
			h = self.h
		else:
			# Quadratic twist will return an isomorphous curve on the
			# GF(sqrt(d)) field extension -> no generator point conversion for
			# now
#			Gx = int(self.G.x * d)
#			Gy = int(self.G.y * d)
			Gx = None
			Gy = None

			# If the original curve had q + 1 - t points, then its twist will
			# have q + 1 + t points. TODO: Does this help us to find the order
			# of the generator point G? I don't think it does :-( Leave n and h
			# therefore unset for the moment.
			n = None
			h = None

		return ShortWeierstrassCurve(a = int(a), b = int(b), p = self.p, n = n, h = h, Gx = Gx, Gy = Gy)

	def twist(self, d = None):
		"""If the twist coefficient d is omitted, the function will
		automatically look for an arbitrary quadratic non-residue in F_P."""
		if d == 0:
			raise Exception("Domain error: d must be nonzero.")
		elif d is None:
			# Search for a QNR in F_P
			d = FieldElement.any_qnr(self.p)
		else:
			d = FieldElement(d, self.p)
			if d.is_qr:
				raise Exception("Twist requested, but twist coefficient d is a quadratic-residue mod p. Refusing to return a GF(p)-isomorphic curve; if you want this behavior, use twist_fp_isomorphic()")
		return self._twist(d)

	def twist_fp_isomorphic(self, u):
		"""Returns a GF(p)-isomorphous curve by applying the substituting
		transformation x = u^2 x' and y = u^3 y' on the curve equation. The
		function therefore returns a quadratic twist with d = u^2, i.e. it
		ensures that the twist coefficient d is a quadratic residue mod p.."""
		if u == 0:
			raise Exception("Domain error: u must be nonzero.")
		return self._twist(FieldElement(u ** 2, self.p), FieldElement(u, self.p))

	def twist_fp_isomorphic_fixed_a(self, a):
		"""Tries to find an GF(p)-isomorphous curve which has a particular
		given value for the curve coefficient 'a'."""

		# anew = a * u^4 -> u = quartic_root(anew / a)
		scalar = a // self.a
		u = scalar.quartic_root()
		if u is None:
			raise NoSuchCurveException("Cannot find an isomorphism so that a = %d because %s has no quartic root in F_P" % (a, scalar))
		return self.twist_fp_isomorphic(int(u))

	def is_isomorphous_curve(self, other):
		"""Returns if the given curve 'other' is isomorphous in the same field
		as the given curve curve."""
		if other.p != self.p:
			return False

		try:
			iso = self.twist_fp_isomorphic_fixed_a(other.a)
		except NoSuchCurveException:
			# No isomorphous curve with this value for a exists
			return False

		# The curves should be identical after the transformation if they're
		# isomorphous to each other
		return (iso.a == other.a) and (iso.b == other.b)

class CurveOpExportSage(object):
	def export_sage(self, varname = "curve"):
		"""Exports the elliptic curve to statements that can be used within the
		SAGE computer algebra system."""

		# EllipticCurve([a1,a2,a3,a4,a6]) means in Sage:
		# y² + a1 x y + a3 y = x³ + a2 x² + a4 x + a6
		# i.e. for Short Weierstrass a4 = A, a6 = B

		statements = [ ]
		statements.append("# %s" % (str(self)))
		statements.append("%s_p = 0x%x" % (varname, int(self.p)))
		statements.append("%s_F = GF(%s_p)" % (varname, varname))
		if self.curvetype == "shortweierstrass":
			statements.append("%s_a = 0x%x" % (varname, int(self.a)))
			statements.append("%s_b = 0x%x" % (varname, int(self.b)))
			statements.append("%s = EllipticCurve(%s_F, [ %s_a, %s_b ])" % (varname, varname, varname, varname))
		else:
			raise Exception(NotImplemented)

		return statements



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import collections
from .FieldElement import FieldElement
from .AffineCurvePoint import AffineCurvePoint
from .EllipticCurve import EllipticCurve
from .DocInherit import doc_inherit
import toyecc.MontgomeryCurve

_TwistedEdwardsCurveDomainParameters = collections.namedtuple("TwistedEdwardsCurveDomainParameters", [ "curvetype", "a", "d", "p", "n", "G" ])

class TwistedEdwardsCurve(EllipticCurve):
	"""Represents an elliptic curve over a finite field F_P that satisfies the
	Twisted Edwards equation a x^2 + y^2 = 1 + d x^2 y^2."""
	pretty_name = "Twisted Edwards"

	def __init__(self, a, d, p, n, h, Gx, Gy, **kwargs):
		"""Create an elliptic Twisted Edwards curve given the equation
		coefficients a and d, the curve field's modulus p, the order of the
		curve n and the generator point G's X and Y coordinates in affine
		representation, Gx and Gy."""
		EllipticCurve.__init__(self, p, n, h, Gx, Gy, **kwargs)
		assert(isinstance(a, int))		# Curve coefficent A
		assert(isinstance(d, int))		# Curve coefficent D
		self._a = FieldElement(a, p)
		self._d = FieldElement(d, p)
		self._name = kwargs.get("name")

		# Check that the curve is not singular
		assert(self.d * (1 - self.d) != 0)

		if self._G is not None:
			# Check that the generator G is on the curve
			assert(self._G.oncurve())

			# Check that the generator G is of curve order
			assert((self.n * self.G).is_neutral)

	@property
	@doc_inherit(EllipticCurve)
	def domainparams(self):
		return _TwistedEdwardsCurveDomainParameters(curvetype = self.curvetype, a = self.a, d = self.d, p = self.p, n = self.n, G = self.G)

	@property
	@doc_inherit(EllipticCurve)
	def curvetype(self):
		return "twistededwards"

	@property
	def a(self):
		"""Returns the coefficient a of the curve equation a x^2 + y^2 = 1 +
		d x^2 y^2."""
		return self._a

	@property
	def d(self):
		"""Returns the coefficient d of the curve equation a x^2 + y^2 = 1 +
		d x^2 y^2."""
		return self._d

	@property
	def B(self):
		"""Returns the length of the curve's field modulus in bits plus one."""
		return self._p.bit_length() + 1

	@property
	def is_complete(self):
		"""Returns if the twisted Edwards curve is complete. This is the case
		exactly when d is a quadratic non-residue modulo p."""
		return self.d.is_qnr

	@doc_inherit(EllipticCurve)
	def neutral(self):
		return AffineCurvePoint(0, 1, self)

	@doc_inherit(EllipticCurve)
	def is_neutral(self, P):
		return (P.x == 0) and (P.y == 1)

	@doc_inherit(EllipticCurve)
	def oncurve(self, P):
		return (self.a * P.x ** 2) + P.y ** 2 == 1 + self.d * P.x ** 2 * P.y ** 2

	@doc_inherit(EllipticCurve)
	def point_conjugate(self, P):
		return AffineCurvePoint(int(-P.x), int(P.y), self)

	@doc_inherit(EllipticCurve)
	def point_addition(self, P, Q):
		x = (P.x * Q.y + Q.x * P.y) // (1 + self.d * P.x * Q.x * P.y * Q.y)
		y = (P.y * Q.y - self.a * P.x * Q.x) // (1 - self.d * P.x * Q.x * P.y * Q.y)
		return AffineCurvePoint(int(x), int(y), self)

	def to_montgomery(self, b = None):
		"""Converts the twisted Edwards curve domain parameters to Montgomery
		domain parameters. For this conversion, b can be chosen semi-freely.
		If the native b coefficient is a quadratic residue modulo p, then the
		freely chosen b value must also be. If it is a quadratic non-residue,
		then so must be the surrogate b coefficient. If b is omitted, the
		native b value is used. The generator point of the twisted Edwards
		curve is also converted to Montgomery form. For this conversion,
		there's an invariant (one of two possible outcomes). An arbitrary
		bijection is used for this."""
		assert((b is None) or isinstance(b, int))

		# Calculate the native montgomery coefficents a, b first
		a = 2 * (self.a + self.d) // (self.a - self.d)
		native_b = 4 // (self.a - self.d)
		if b is None:
			b = native_b
		else:
			# If a b value was supplied, make sure is is either a QR or QNR mod
			# p, depending on what the native b value was
			b = FieldElement(b, self.p)
			if native_b.is_qr != b.is_qr:
				raise Exception("The b coefficient of the resulting curve must be a quadratic %s modulo p, %s is not." % ([ "non-residue", "residue" ][native_b.is_qr], str(b)))

		# Generate the raw curve without a generator yet
		raw_curve = toyecc.MontgomeryCurve.MontgomeryCurve(
			a = int(a),
			b = int(b),
			p = self.p,
			n = self.n,
			h = self.h,
			Gx = None,
			Gy = None,
		)

		# Then convert the original generator point using the raw curve to
		# yield a birationally equivalent generator point
		G_m = self.G.convert(raw_curve)

		# And create the curve again, setting this generator
		montgomery_curve = toyecc.MontgomeryCurve.MontgomeryCurve(
			a = int(a),
			b = int(b),
			p = self.p,
			n = self.n,
			h = self.h,
			Gx = int(G_m.x),
			Gy = int(G_m.y),
		)

		return montgomery_curve

	def __str__(self):
		if self.hasname:
			return "TwistedEdwardsCurve<%s>" % (self.name)
		else:
			return "TwistedEdwardsCurve<0x%x x^2 + y^2 = 1 + 0x%x x^2 y^2 mod 0x%x>" % (int(self.a), int(self.d), int(self.p))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
import random
from .. import getcurvebyname

class TwEdMontDomainTests(unittest.TestCase):
	def setUp(self):
		self._mont = getcurvebyname("curve25519")
		self._twed = getcurvebyname("ed25519")

	def test_twed_to_mont(self):
		mont = self._twed.to_montgomery(b = int(self._mont.b))
		self.assertEqual(mont.domainparams, self._mont.domainparams)

	def test_mont_to_twed(self):
		twed = self._mont.to_twistededwards(a = int(self._twed.a))
		self.assertEqual(twed.domainparams, self._twed.domainparams)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
from .. import FieldElement, getcurvebyname, ECPublicKey, ECPrivateKey

class Ed448BasicTests(unittest.TestCase):
	def test_sign_verify(self):
		curves = [getcurvebyname("Ed448-Goldilocks"), getcurvebyname("Ed448")]
		for curve in curves:
			privkey = ECPrivateKey.eddsa_generate(curve)

			msg = b"foobar"
			signature = privkey.eddsa_sign(msg)

			self.assertTrue(privkey.pubkey.eddsa_verify(msg, signature))
			self.assertFalse(privkey.pubkey.eddsa_verify(msg + b"x", signature))

	def test_sig_encode_decode(self):
		curves = [ getcurvebyname("Ed448"), getcurvebyname("Ed448-Goldilocks") ]
		for curve in curves:
			privkey = ECPrivateKey.eddsa_generate(curve)
			msg = b"foobar"
			signature = privkey.eddsa_sign(msg)

			encoded_signature = signature.encode()
			decoded_signature = ECPrivateKey.EDDSASignature.decode(curve, encoded_signature)
			self.assertEqual(decoded_signature, signature)
			self.assertTrue(privkey.pubkey.eddsa_verify(msg, signature))
			self.assertTrue(privkey.pubkey.eddsa_verify(msg, decoded_signature))

	def test_seeding_signing(self):
		curve = getcurvebyname("Ed448-Goldilocks")
		seed = bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b")

		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 521658399617511624509929819094270498323007786671637499019582168374758478770958028340603419308639592898868374490003595203618871291427304)

		pubkey = privkey.pubkey
		self.assertEqual(pubkey.point.x, 0xf6882451bdd0174e32c5b38a637659cd839ef9cc40b53843adda3b01886a86edc71e8af14839b8bb21e185775ed3f61a105d0bf339d04ac7)
		self.assertEqual(pubkey.point.y, 0xca023dfc9ed27809e1ca6976cb18618cf066bdd0769dd8962bac9c9535c7bf092090dfa289a997a332ed9c3984ed085dbf9dce6effb489d6)

		msg = b"Foobar!"
		signature = privkey.eddsa_sign(msg)

		self.assertEqual(signature.R.x, 0x6e9acf5065ca854d5d492b85e9178cbad79646ad450eee2a033b8ae7185b8baee834d37726964954cc66d732c52748e3dfdd63d032960e98)
		self.assertEqual(signature.R.y, 0x1d460c01a5a31d3ee5271be4295464cfcb1b8354c6ba09e95c9d440b6eb0cfa3733c40bff6b8384b3eec45686d222e511e6b133db4203b10)
		self.assertEqual(signature.s, 0x325c4c3d9fe2873c6ba8e403abe00543d9b2245599036e32fc10b83e86019bb8bacb78ded80b70a923d2fa1a7ca606639418082cb352014c)

	def test_key_encoding(self):
		curve = getcurvebyname("Ed448-Goldilocks")
		seed = bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b")

		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)

		self.assertEqual(privkey.eddsa_encode(), bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b"))
		self.assertEqual(privkey.pubkey.eddsa_encode(), bytes.fromhex("d689b4ff6ece9dbf5d08ed84399ced32a397a989a2df902009bfc735959cac2b96d89d76d0bd66f08c6118cb7669cae10978d29efc3d02ca80"))

	def test_key_decoding(self):
		curve = getcurvebyname("Ed448-Goldilocks")
		seed = bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b")

		privkey = ECPrivateKey.eddsa_decode(curve, bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b"))
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 521658399617511624509929819094270498323007786671637499019582168374758478770958028340603419308639592898868374490003595203618871291427304)

		pubkey = ECPublicKey.eddsa_decode(curve, bytes.fromhex("71c7391d7a5df8dc1aa14fae59256e7bbf4b4f12065a435f7fe0b36a71de0db7e09d32074ad2baa5773f16f85c1171945912f5b08a0990b900"))

	def test_keys_ed448_rfc8032(self):
		curve = getcurvebyname("Ed448")

		# RFC8032 testvector 1
		seed = bytes.fromhex("6c82a562cb808d10d632be89c8513ebf6c929f34ddfa8c9f63c9960ef6e348a3528c8a3fcc2f044e39a3fc5b94492f8f032e7549a20098f95b")
		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 521658399617511624509929819094270498323007786671637499019582168374758478770958028340603419308639592898868374490003595203618871291427304)

		pubkey = privkey.pubkey
		self.assertEqual(pubkey.point.x, 0xb02f7d0580fd8e88f3fc8ecd47f43499b0000faf1e84d0c2283736c991c4a64447ce4e8d8a6c74010baf726ef20006bcf1fa990e7a822287)
		self.assertEqual(pubkey.point.y, 0x6125e8afbe1afad16c0fe5f13d78d61b06c7469b7624f1ed7867e9805da70e8a1f0ea7852434a11d6ad46a61ec87e72cfd61b4599b44d75f)
		self.assertEqual(pubkey.eddsa_encode(), bytes.fromhex("5fd7449b59b461fd2ce787ec616ad46a1da1342485a70e1f8a0ea75d80e96778edf124769b46c7061bd6783df1e50f6cd1fa1abeafe8256180"))

		msg = b"Foobar!"
		signature = privkey.eddsa_sign(msg)

		self.assertEqual(signature.R.x, 0xe7ff455957ad25768b75c9257fd9757706aa473061636eebfb24cc54fd944fe7e2aa8b07283a30e8c2d96a357b17c7e6ba83f6ece4f0ff17)
		self.assertEqual(signature.R.y, 0xc49a90fb910f4e6eb4a25ad3dc76d8e2f273d8a3b738374dc1310e7b6fd71185d96aa47c2436b63a64924da58879ccdb9437e026521692be)
		self.assertEqual(signature.s, 0x37754cfdfce43e25c7a49b7cbd18bbbfa2746e154d5f82a00ae274dcc5c88329eb7e51f6ccdc29cc4cf0d4bc05ad5a4c8f39ce6275dbbf4)

		# RFC8032 testvector 2
		seed = bytes.fromhex("c4eab05d357007c632f3dbb48489924d552b08fe0c353a0d4a1f00acda2c463afbea67c5e8d2877c5e3bc397a659949ef8021e954e0a12274e")
		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 689909312335174883516499642766841446935258538413003095779028193821563285076416857945471525085401816906939882844448039967110836770482568)

		pubkey = privkey.pubkey
		self.assertEqual(pubkey.point.x, 0xd2ffd9b53d3bff141f68cf0248020a14e2f1aba433258c33c290607f78eab2c6b969bdf226775ba7db74bcf20ffddab9d3ada6b23a0ee857)
		self.assertEqual(pubkey.point.y, 0x943a4c7b626051239c1682cba48e43b802287400eb01ea6a86c098676c0cfa2b37c058935da534c80acd7e5f5431e56a45ffcd30f428ba43)
		self.assertEqual(pubkey.eddsa_encode(), bytes.fromhex("43ba28f430cdff456ae531545f7ecd0ac834a55d9358c0372bfa0c6c6798c0866aea01eb00742802b8438ea4cb82169c235160627b4c3a9480"))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
import tempfile
import uuid
import os
from ..ECPrivateKey import ECPrivateKey
from ..ECPublicKey import ECPublicKey
from ..Exceptions import NoSuchCurveException, UnsupportedFieldException
from ..ASN1 import have_asn1_support
from .. import getcurvebyname

class _Tempfile(object):
	def __init__(self, content):
		self._content = content
		self._filename = None

	def __enter__(self):
		assert(self._filename is None)
		self._filename = tempfile.gettempdir() + "/tmp_toyecc_" + str(uuid.uuid4())
		with open(self._filename, "w") as f:
			print(self._content, file = f)
		return self._filename

	def __exit__(self, *args):
		assert(self._filename is not None)
		os.unlink(self._filename)


@unittest.skipIf(not have_asn1_support(), "ASN.1 support not available")
class KeyLoadStoreTests(unittest.TestCase):
	_PRIVKEY_PEM_OID = _Tempfile("""-----BEGIN EC PARAMETERS-----
BgUrgQQAHQ==
-----END EC PARAMETERS-----
-----BEGIN EC PRIVATE KEY-----
MEQCAQEEEArzxc2vyxUaQdXEj9bdkHmgBwYFK4EEAB2hJAMiAATKd4WyFLzs86V0
TTQRwZ1rda0gr5vCWg+t15qmrmSevA==
-----END EC PRIVATE KEY-----
""")

	_PRIVKEY_PEM_EXPLICIT = _Tempfile("""-----BEGIN EC PRIVATE KEY-----
MIHXAgEBBBAK88XNr8sVGkHVxI/W3ZB5oIGZMIGWAgEBMBwGByqGSM49AQECEQD/
///9////////////////MDsEENYDGZjRs7v+v1nMm7/5ruEEEF7u/KOA0CkZ3Cxl
WLttil0DFQAATWluZ2h1YVF1EtjwNDH85juI9AQhBHtqpdheVymD5vsyp83rwUAn
tpFqiU067nEG/oBfw0tEAhA/////f////74AJHIGE7WjAgEEoSQDIgAEyneFshS8
7POldE00EcGda3WtIK+bwloPrdeapq5knrw=
-----END EC PRIVATE KEY-----
""")

	_PRIVKEY_PEM_F2M_OID = _Tempfile("""-----BEGIN EC PRIVATE KEY-----
MEcCAQEEEQFnPiX+TOQzUFepnIjE1d3QoAcGBSuBBAAWoSYDJAAEAtQX7xbkBCiu
HuYYzjcKY64EuBxr4Es8/Is+UnZZ3oV+zQ==
-----END EC PRIVATE KEY-----
""")

	_PRIVKEY_PEM_F2M_EXPLICIT = _Tempfile("""-----BEGIN EC PRIVATE KEY-----
MIHoAgEBBBEBZz4l/kzkM1BXqZyIxNXd0KCBpzCBpAIBATAlBgcqhkjOPQECMBoC
AgCDBgkqhkjOPQECAwMwCQIBAgIBAwIBCDA9BBEHoRsJp2tWIURBj/P/jCVwuAQR
AhfAVhCIS2O5xscpFnj500EDFQBNaW5naHVhUXWYW9OtutohtDqX4gQjBACBuvkf
35gzxA+cGBNDY4OZB4xufqOMAB9zyBNLG0754VACEQQAAAAAAAAAAjEjlTqUZLVN
AgECoSYDJAAEAtQX7xbkBCiuHuYYzjcKY64EuBxr4Es8/Is+UnZZ3oV+zQ==
-----END EC PRIVATE KEY-----
""")

	_PUBKEY_PEM_OID = _Tempfile("""-----BEGIN PUBLIC KEY-----
MDIwEAYHKoZIzj0CAQYFK4EEAAYDHgAEF+RQ5RcwZ36icW199nTPE8AvIMZr4uRk
vKaJtw==
-----END PUBLIC KEY-----
""")

	_PUBKEY_PEM_EXPLICIT = _Tempfile("""-----BEGIN PUBLIC KEY-----
MIG6MIGXBgcqhkjOPQIBMIGLAgEBMBoGByqGSM49AQECDwDbfCq/YuNeZoB2vq0g
izA3BA7bfCq/YuNeZoB2vq0giAQOZZ74ugQ5Fu7eiRFwKyIDFQAA9QsCjk1pbmdo
dWFRdSkEcng/sQQdBAlIcjmZWl7na1X5wvCYqJzlr4ckwKI+Dg/3dQACDwDbfCq/
YuNedijfrGVhxQIBAQMeAAQX5FDlFzBnfqJxbX32dM8TwC8gxmvi5GS8pom3
-----END PUBLIC KEY-----
""")


	def test_load_privkey_fp_oid(self):
		with self._PRIVKEY_PEM_OID as filename:
			key = ECPrivateKey.load_pem(filename)
		self.assertEqual(key.scalar, 0x0af3c5cdafcb151a41d5c48fd6dd9079)
		self.assertEqual(key.pubkey.point.x, 0xca7785b214bcecf3a5744d3411c19d6b)
		self.assertEqual(key.pubkey.point.y, 0x75ad20af9bc25a0fadd79aa6ae649ebc)

	def test_load_privkey_fp_explicit(self):
		with self._PRIVKEY_PEM_EXPLICIT as filename:
			key = ECPrivateKey.load_pem(filename)
		self.assertEqual(key.scalar, 0x0af3c5cdafcb151a41d5c48fd6dd9079)
		self.assertEqual(key.pubkey.point.x, 0xca7785b214bcecf3a5744d3411c19d6b)
		self.assertEqual(key.pubkey.point.y, 0x75ad20af9bc25a0fadd79aa6ae649ebc)
		self.assertEqual(key.curve, getcurvebyname("secp128r2"))

	def test_load_privkey_f2m_oid(self):
		with self._PRIVKEY_PEM_F2M_OID as filename:
			with self.assertRaises(NoSuchCurveException):
				# No such OID in database
				ECPrivateKey.load_pem(filename)

	def test_load_privkey_f2m_explicit(self):
		with self._PRIVKEY_PEM_F2M_EXPLICIT as filename:
			with self.assertRaises(UnsupportedFieldException):
				# Unsupported field F(2^m)
				ECPrivateKey.load_pem(filename)

	def test_load_pubkey_fp_oid(self):
		with self._PUBKEY_PEM_OID as filename:
			key = ECPublicKey.load_pem(filename)
			self.assertEqual(key.point.x, 0x17e450e51730677ea2716d7df674)
			self.assertEqual(key.point.y, 0xcf13c02f20c66be2e464bca689b7)
			self.assertEqual(key.point.curve, getcurvebyname("secp112r1"))

	def test_load_pubkey_fp_explicit(self):
		with self._PUBKEY_PEM_EXPLICIT as filename:
			key = ECPublicKey.load_pem(filename)
			self.assertEqual(key.point.x, 0x17e450e51730677ea2716d7df674)
			self.assertEqual(key.point.y, 0xcf13c02f20c66be2e464bca689b7)
			self.assertEqual(key.point.curve, getcurvebyname("secp112r1"))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import hashlib

from .FieldElement import FieldElement
from .AffineCurvePoint import AffineCurvePoint
from .Random import secure_rand_int_between
from . import Tools
from .PrivKeyOps import PrivKeyOpLoad
from .ASN1 import parse_asn1_public_key
from .CurveDB import CurveDB
from .CurveQuirks import CurveQuirkSigningHashFunction

class PubKeyOpECDSAExploitReusedNonce(object):
	def ecdsa_exploit_reused_nonce(self, msg1, sig1, msg2, sig2):
		"""Given two different messages msg1 and msg2 and their corresponding
		signatures sig1, sig2, try to calculate the private key that was used
		for signing if during signature generation no unique nonces were
		used."""
		assert(isinstance(msg1, bytes))
		assert(isinstance(msg2, bytes))
		assert(msg1 != msg2)
		assert(sig1.r == sig2.r)

		# Hash the messages
		dig1 = hashlib.new(sig1.hashalg)
		dig1.update(msg1)
		dig1 = dig1.digest()
		dig2 = hashlib.new(sig2.hashalg)
		dig2.update(msg2)
		dig2 = dig2.digest()

		# Calculate hashes of messages
		e1 = Tools.ecdsa_msgdigest_to_int(dig1, self.point.curve.n)
		e2 = Tools.ecdsa_msgdigest_to_int(dig2, self.point.curve.n)

		# Take them modulo n
		e1 = FieldElement(e1, self.point.curve.n)
		e2 = FieldElement(e2, self.point.curve.n)

		(s1, s2) = (FieldElement(sig1.s, self.point.curve.n), FieldElement(sig2.s, self.point.curve.n))
		r = sig1.r

		# Recover (supposedly) random nonce
		nonce = (e1 - e2) // (s1 - s2)

		# Recover private key
		priv = ((nonce * s1) - e1) // r

		return { "nonce": nonce, "privatekey": priv }


class PubKeyOpECDSAVerify(object):
	def ecdsa_verify_hash(self, message_digest, signature):
		"""Verify ECDSA signature over the hash of a message (the message
		digest)."""
		assert(isinstance(message_digest, bytes))
		assert(0 < signature.r < self.curve.n)
		assert(0 < signature.s < self.curve.n)

		# Convert message digest to integer value
		e = Tools.ecdsa_msgdigest_to_int(message_digest, self.curve.n)

		(r, s) = (signature.r, FieldElement(signature.s, self.curve.n))
		w = s.inverse()
		u1 = int(e * w)
		u2 = int(r * w)

		pt = (u1 * self.curve.G) + (u2 * self.point)
		x1 = int(pt.x) % self.curve.n
		return x1 == r

	def ecdsa_verify(self, message, signature):
		"""Verify an ECDSA signature over a message."""
		assert(isinstance(message, bytes))
		digest_fnc = hashlib.new(signature.hashalg)
		digest_fnc.update(message)
		message_digest = digest_fnc.digest()
		return self.ecdsa_verify_hash(message_digest, signature)


class PubKeyOpEDDSAVerify(object):
	def eddsa_verify(self, message, signature):
		"""Verify an EdDSA signature over a message."""
		if not self.curve.has_quirk(CurveQuirkSigningHashFunction):
			raise Exception("Unable to determine EdDSA signature function.")
		quirk = self.curve.get_quirk(CurveQuirkSigningHashFunction)
		h = Tools.bytestoint_le(quirk.hashdata(signature.R.eddsa_encode() + self.point.eddsa_encode() + message))
		return (signature.s * self.curve.G) == signature.R + (h * self.point)


class PubKeyOpEDDSAEncode(object):
	def eddsa_encode(self):
		"""Encodes a EdDSA-encoded public key to its serialized (bytes)
		form."""
		return self.point.eddsa_encode()

	@classmethod
	def eddsa_decode(cls, curve, encoded_pubkey):
		"""Decodes a EdDSA-encoded public key from its serialized (bytes)
		form."""
		pubkey = AffineCurvePoint.eddsa_decode(curve, encoded_pubkey)
		return cls(pubkey)

class PubKeyOpECIESEncrypt(object):
	def ecies_encrypt(self, r = None):
		"""Generates a shared secret which can be used to symetrically encrypt
		data that only the holder of the corresponding private key can read.
		The output are two points, R and S: R is the public point that is
		transmitted together with the message while S is the point which
		resembles the shared secret. The receiver can use R together with her
		private key to reconstruct S. A random nonce r can be supplied for this
		function. If it isn't supplied, it is randomly chosen."""

		# Chose a random number
		if r is None:
			r = secure_rand_int_between(1, self.curve.n - 1)

		R = r * self.curve.G
		S = r * self.point

		# Return the publicly transmitted R and the symmetric key S
		return { "R": R, "S": S }


class PubKeyOpLoad(object):
	@classmethod
	def load_derdata(cls, derdata):
		"""Loads an EC public key from a DER-encoded ASN.1 bytes object."""
		asn1 = parse_asn1_public_key(derdata)
		curve = CurveDB().get_curve_from_asn1(asn1["algorithm"]["parameters"])
		point = AffineCurvePoint.deserialize_uncompressed(Tools.bits_to_bytes(asn1["subjectPublicKey"]), curve)
		return cls(point)

	@classmethod
	def load_pem(cls, pemfilename):
		"""Loads an EC public key from a PEM-encoded 'PUBLIC KEY' file."""
		return cls.load_derdata(Tools.load_pem_data(pemfilename, "PUBLIC KEY"))

	@classmethod
	def load_der(cls, derfilename):
		"""Loads an EC public key from a DER-encoded ASN.1 file."""
		with open(derfilename, "rb") as f:
			data = f.read()
			return cls.load_derdata(data)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

from .FieldElement import FieldElement

class CRT():
	"""Implements the Chinese Remainder Theorem algorithm where a number of
	modular congruences are given that all need to be satisfied."""
	def __init__(self):
		self._moduli = { }

	def add(self, value, modulus):
		"""Adds a value that shall be returned when the result is taken modulo
		the given modulus."""
		assert(modulus not in self._moduli)
		assert(isinstance(value, int))
		assert(isinstance(modulus, int))
		self._moduli[modulus] = value
		return self

	def solve(self):
		"""Solve the Chinese Remainder Theorem for the given values and
		moduli."""
		# Calculate product of all moduli
		product = 1
		for modulus in self._moduli.keys():
			product *= modulus

		# Then determine the solution
		solution = 0
		for modulus in self._moduli.keys():
			if self._moduli[modulus] == 0:
				continue

			rem_product = product // modulus
			one_value = int(FieldElement(rem_product, modulus).inverse())
			solution += rem_product * one_value * self._moduli[modulus]

		return solution % product



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import enum
import collections
from .ShortWeierstrassCurve import ShortWeierstrassCurve
from .MontgomeryCurve import MontgomeryCurve
from .TwistedEdwardsCurve import TwistedEdwardsCurve
from .Singleton import singleton
from .FieldElement import FieldElement
from .Exceptions import DuplicateCurveException, NoSuchCurveException, UnsupportedFieldException
from .ASN1 import parse_asn1_field_params_fp
from .AffineCurvePoint import AffineCurvePoint
from .CurveQuirks import CurveQuirkEdDSASetPrivateKeyMSB, CurveQuirkEdDSAEnsurePrimeOrderSubgroup, CurveQuirkSigningHashFunction
from . import Tools

@singleton
class CurveDB(object):
	def __init__(self):
		self._entries = { }
		self._primary_names = set()
		self._taken_names = set()

	def _checknames(self, curvenames):
		if len(curvenames & self._taken_names) > 0:
			taken_names = ", ".join(sorted(list(curvenames & self._taken_names)))
			raise DuplicateCurveException("Curve(s) named %s already registered in curve DB." % (taken_names))

	def register(self, entry):
		"""Registers a curve in the curve database."""
		all_names = set(name.lower() for name in entry.all_aliases)
		self._checknames(all_names)
		self._taken_names |= all_names
		self._primary_names.add(entry.name)

		self._entries[entry.primary_name.lower()] = entry
		for aliasname in entry.aliases:
			clone = entry.clone(secondary_name = aliasname)
			self._entries[aliasname.lower()] = clone

	def curvenames(self):
		"""Returns the primary names of all curves in the DB."""
		return (curve.name for curve in self._entries.values() if (curve.is_aka is False))

	def allcurvenames(self):
		"""Returns all names of all curves in the DB. This includes duplicate
		AKAs such as secp224r1 which is also known as wap-wsg-idm-ecid-wtls12
		albeit under a different OID."""
		return (curve.name for curve in self._entries.values())

	def find_duplicate_curves(self):
		"""Returns curves in which the domain parameters (including the
		coordinates of the generator point G) are identical. This can happen if
		identical curves are registered under the same name."""
		params = collections.defaultdict(list)
		for curve in self:
			params[tuple(sorted(curve.domain_params))].append(curve.name)
		return [ curves for (param, curves) in params.items() if (len(curves) > 1) ]

	def getentry(self, name):
		"""Returns a specific curve entry by its case-insensitive name."""
		name = name.lower()
		if name not in self._entries:
			raise KeyError("Curve named '%s' is not known in curve database." % (name))
		return self._entries[name]

	def get_curve_from_asn1(self, asn1):
		"""This function will take a parsed ASN.1 ECParameters class as input
		and try to return the curve specified within. If the ECParameters
		specify a named curve by its's OID then a lookup is performed against
		the curve database and that named curve returned on success if
		non-ambiguous. If the parameters are exclicitly stated, then an unnamed
		ShortWeierstrassCurve is constructed."""

		if asn1["namedCurve"] is not None:
			# Curve is encoded as OID, look up from curve DB
			curve_oid = str(asn1["namedCurve"])
			entries = [ entry for entry in self if (entry.oid == curve_oid) ]
			if len(entries) == 0:
				raise NoSuchCurveException("Trying to load curve with OID %s from curve DB, but no such curve is present in database." % (curve_oid))
			elif len(entries) > 1:
				raise Exception("Trying to load curve with OID %s from curve DB, but found %d curves (refuse to guess in the face of ambiguity)." % (curve_oid, len(entries)))
			curve = entries[0]()
		elif asn1["specifiedCurve"] is not None:
			field_type_oid = str(asn1["specifiedCurve"]["fieldID"]["fieldType"])
			if field_type_oid == "1.2.840.10045.1.1":
				# F_P curve is encoded in explicit form
				p = int(parse_asn1_field_params_fp(asn1["specifiedCurve"]["fieldID"]["parameters"]))
				a = Tools.bytestoint(asn1["specifiedCurve"]["curve"]["a"])
				b = Tools.bytestoint(asn1["specifiedCurve"]["curve"]["b"])
				G = bytes(asn1["specifiedCurve"]["base"])
				(Gx, Gy) = AffineCurvePoint.deserialize_uncompressed(G)
				n = int(asn1["specifiedCurve"]["order"])
				h = int(asn1["specifiedCurve"]["cofactor"])
				curve = ShortWeierstrassCurve(p = p, a = a, b = b, n = n, h = h, Gx = Gx, Gy = Gy)
			else:
				# Maybe F_2^N curve or some other, unsupported type
				raise UnsupportedFieldException("Only supports elliptic curves in F_P are supported, but the requested field type OID was %s." % (field_type_oid))
		else:
			raise NoSuchCurveException("Cannot load implicit curve.")
		return curve

	def __iter__(self):
		"""Iterates over the curve DB entries."""
		for name in self.curvenames():
			yield self._entries[name.lower()]

	def __getitem__(self, name):
		"""Returns a curve (not a curve DB entry) by its name."""
		return self.getentry(name)()

	def __str__(self):
		return "CurveDB<%d unique curves, %d total>" % (len(self._primary_names), len(self._entries))


class _CurveDBEntry(object):
	def __init__(self, primary_name, curve_class, domain_params, **kwargs):
		allowed_kwargs = set(("oid", "alt_oids", "aliases", "origin", "secure", "quirks"))
		wrong_args = kwargs.keys() - allowed_kwargs
		if len(wrong_args) > 0:
			raise Exception("Illegal keyword arguments: %s" % (", ".join(sorted(wrong_args))))

		assert(primary_name is not None)
		self._primary_name = primary_name
		self._secondary_name = None
		self._curve_class = curve_class
		self._domain_params = domain_params
		self._oid = kwargs.get("oid")
		self._alt_oids = kwargs.get("alt_oids")
		self._aliases = kwargs.get("aliases")
		self._origin = kwargs.get("origin")
		self._secure = kwargs.get("secure", True)
		self._quirks = kwargs.get("quirks", [ ])
		self._instance = None

	def clone(self, secondary_name = None):
		clone = _CurveDBEntry(primary_name = self._primary_name, curve_class = self._curve_class, domain_params = self._domain_params, oid = self._oid, alt_oids = self._alt_oids, aliases = self._aliases, origin = self._origin, secure = self._secure)
		clone._instance = self._instance
		clone._secondary_name = secondary_name
		return clone

	@property
	def is_aka(self):
		"""Returns if this curve entry is an AKA ('also known as') for a
		different curve (but maybe with a different OID). Example: prime192v1
		and secp192r1 refer to identical curves, but 'prime192v1' is the
		internally considered primary name while 'secp192r1' is considered to
		be an AKA."""
		return self._secondary_name is not None

	@property
	def primary_name(self):
		return self._primary_name

	@property
	def name(self):
		if self._secondary_name is not None:
			return self._secondary_name
		else:
			return self._primary_name

	@property
	def fieldsize_bits(self):
		return self._domain_params["p"].bit_length()

	@property
	def secure(self):
		return self._secure

	@property
	def origin(self):
		return self._origin

	@property
	def bits_security_estimate(self):
		if not self.secure:
			return 0
		else:
			# Require instanciation of the class
			self()
			return self._instance.security_bit_estimate

	def get_alternative_oid(self, name):
		"""Returns the alternative OID if it has one."""
		if self._alt_oids is not None:
			return self._alt_oids.get(name)

	@property
	def oid(self):
		if (self._alt_oids is not None) and (self.name in self._alt_oids):
			return self._alt_oids[self.name]
		else:
			return self._oid

	@property
	def aliases(self):
		if self._aliases is not None:
			yield from self._aliases

	@property
	def all_aliases(self):
		yield self._primary_name
		yield from self.aliases

	@property
	def prettyname(self):
		if self._instance is None:
			return self._curve_class.pretty_name
		else:
			return self._instance.prettyname

	@property
	def domain_params(self):
		if self._instance is None:
			return dict(self._domain_params)
		else:
			return self._instance.domainparamdict

	@property
	def prettytitle(self):
		return "%d bit %s Curve" % (self.fieldsize_bits, self.prettyname)

	def dump(self, domain = False):
		print("%s: %s" % (self.name, self.prettytitle))
		if self._aliases is not None:
			print("Aliases: %s" % (", ".join(sorted(list(self._aliases)))))
		if self._oid is not None:
			print("OID    : %s" % (self._oid))
		if domain:
			print("Domain parameters:")
			for (key, value) in sorted(self.domain_params.items()):
				if isinstance(value, FieldElement):
					value = value.sigint()
				print("    %-10s %s" % (key, value))

	def __call__(self):
		"""Instanciate the curve."""
		if self._instance is None:
			# Instanciate actual curve
			params = self._domain_params
			params["name"] = self.name
			params["quirks"] = self._quirks
			self._instance = self._curve_class(**params)
		return self._instance

	def __str__(self):
		if self._secondary_name is not None:
			return "CurveDBEntry<%s AKA %s>" % (self.primary_name, self._secondary_name)
		else:
			return "CurveDBEntry<%s>" % (self.name)

cdb = CurveDB()
cdb.register(_CurveDBEntry("brainpoolP160r1", ShortWeierstrassCurve, {
	"a": 0x340e7be2a280eb74e2be61bada745d97e8f7c300,
	"b": 0x1e589a8595423412134faa2dbdec95c8d8675e58,
	"p": 0xe95e4a5f737059dc60dfc7ad95b3d8139515620f,
	"n": 0xe95e4a5f737059dc60df5991d45029409e60fc09,
	"h": 1,
	"Gx": 0xbed5af16ea3f6a4f62938c4631eb5af7bdbcdbc3,
	"Gy": 0x1667cb477a1a8ec338f94741669c976316da6321,
}, oid = "1.3.36.3.3.2.8.1.1.1", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP160t1", ShortWeierstrassCurve, {
	"a": 0xe95e4a5f737059dc60dfc7ad95b3d8139515620c,
	"b": 0x7a556b6dae535b7b51ed2c4d7daa7a0b5c55f380,
	"p": 0xe95e4a5f737059dc60dfc7ad95b3d8139515620f,
	"n": 0xe95e4a5f737059dc60df5991d45029409e60fc09,
	"h": 1,
	"Gx": 0xb199b13b9b34efc1397e64baeb05acc265ff2378,
	"Gy": 0xadd6718b7c7c1961f0991b842443772152c9e0ad,
}, oid = "1.3.36.3.3.2.8.1.1.2", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP192r1", ShortWeierstrassCurve, {
	"a": 0x6a91174076b1e0e19c39c031fe8685c1cae040e5c69a28ef,
	"b": 0x469a28ef7c28cca3dc721d044f4496bcca7ef4146fbf25c9,
	"p": 0xc302f41d932a36cda7a3463093d18db78fce476de1a86297,
	"n": 0xc302f41d932a36cda7a3462f9e9e916b5be8f1029ac4acc1,
	"h": 1,
	"Gx": 0xc0a0647eaab6a48753b033c56cb0f0900a2f5c4853375fd6,
	"Gy": 0x14b690866abd5bb88b5f4828c1490002e6773fa2fa299b8f,
}, oid = "1.3.36.3.3.2.8.1.1.3", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP192t1", ShortWeierstrassCurve, {
	"a": 0xc302f41d932a36cda7a3463093d18db78fce476de1a86294,
	"b": 0x13d56ffaec78681e68f9deb43b35bec2fb68542e27897b79,
	"p": 0xc302f41d932a36cda7a3463093d18db78fce476de1a86297,
	"n": 0xc302f41d932a36cda7a3462f9e9e916b5be8f1029ac4acc1,
	"h": 1,
	"Gx": 0x3ae9e58c82f63c30282e1fe7bbf43fa72c446af6f4618129,
	"Gy": 0x97e2c5667c2223a902ab5ca449d0084b7e5b3de7ccc01c9,
}, oid = "1.3.36.3.3.2.8.1.1.4", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP224r1", ShortWeierstrassCurve, {
	"a": 0x68a5e62ca9ce6c1c299803a6c1530b514e182ad8b0042a59cad29f43,
	"b": 0x2580f63ccfe44138870713b1a92369e33e2135d266dbb372386c400b,
	"p": 0xd7c134aa264366862a18302575d1d787b09f075797da89f57ec8c0ff,
	"n": 0xd7c134aa264366862a18302575d0fb98d116bc4b6ddebca3a5a7939f,
	"h": 1,
	"Gx": 0xd9029ad2c7e5cf4340823b2a87dc68c9e4ce3174c1e6efdee12c07d,
	"Gy": 0x58aa56f772c0726f24c6b89e4ecdac24354b9e99caa3f6d3761402cd,
}, oid = "1.3.36.3.3.2.8.1.1.5", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP224t1", ShortWeierstrassCurve, {
	"a": 0xd7c134aa264366862a18302575d1d787b09f075797da89f57ec8c0fc,
	"b": 0x4b337d934104cd7bef271bf60ced1ed20da14c08b3bb64f18a60888d,
	"p": 0xd7c134aa264366862a18302575d1d787b09f075797da89f57ec8c0ff,
	"n": 0xd7c134aa264366862a18302575d0fb98d116bc4b6ddebca3a5a7939f,
	"h": 1,
	"Gx": 0x6ab1e344ce25ff3896424e7ffe14762ecb49f8928ac0c76029b4d580,
	"Gy": 0x374e9f5143e568cd23f3f4d7c0d4b1e41c8cc0d1c6abd5f1a46db4c,
}, oid = "1.3.36.3.3.2.8.1.1.6", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP256r1", ShortWeierstrassCurve, {
	"a": 0x7d5a0975fc2c3057eef67530417affe7fb8055c126dc5c6ce94a4b44f330b5d9,
	"b": 0x26dc5c6ce94a4b44f330b5d9bbd77cbf958416295cf7e1ce6bccdc18ff8c07b6,
	"p": 0xa9fb57dba1eea9bc3e660a909d838d726e3bf623d52620282013481d1f6e5377,
	"n": 0xa9fb57dba1eea9bc3e660a909d838d718c397aa3b561a6f7901e0e82974856a7,
	"h": 1,
	"Gx": 0x8bd2aeb9cb7e57cb2c4b482ffc81b7afb9de27e1e3bd23c23a4453bd9ace3262,
	"Gy": 0x547ef835c3dac4fd97f8461a14611dc9c27745132ded8e545c1d54c72f046997,
}, oid = "1.3.36.3.3.2.8.1.1.7", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP256t1", ShortWeierstrassCurve, {
	"a": 0xa9fb57dba1eea9bc3e660a909d838d726e3bf623d52620282013481d1f6e5374,
	"b": 0x662c61c430d84ea4fe66a7733d0b76b7bf93ebc4af2f49256ae58101fee92b04,
	"p": 0xa9fb57dba1eea9bc3e660a909d838d726e3bf623d52620282013481d1f6e5377,
	"n": 0xa9fb57dba1eea9bc3e660a909d838d718c397aa3b561a6f7901e0e82974856a7,
	"h": 1,
	"Gx": 0xa3e8eb3cc1cfe7b7732213b23a656149afa142c47aafbc2b79a191562e1305f4,
	"Gy": 0x2d996c823439c56d7f7b22e14644417e69bcb6de39d027001dabe8f35b25c9be,
}, oid = "1.3.36.3.3.2.8.1.1.8", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP320r1", ShortWeierstrassCurve, {
	"a": 0x3ee30b568fbab0f883ccebd46d3f3bb8a2a73513f5eb79da66190eb085ffa9f492f375a97d860eb4,
	"b": 0x520883949dfdbc42d3ad198640688a6fe13f41349554b49acc31dccd884539816f5eb4ac8fb1f1a6,
	"p": 0xd35e472036bc4fb7e13c785ed201e065f98fcfa6f6f40def4f92b9ec7893ec28fcd412b1f1b32e27,
	"n": 0xd35e472036bc4fb7e13c785ed201e065f98fcfa5b68f12a32d482ec7ee8658e98691555b44c59311,
	"h": 1,
	"Gx": 0x43bd7e9afb53d8b85289bcc48ee5bfe6f20137d10a087eb6e7871e2a10a599c710af8d0d39e20611,
	"Gy": 0x14fdd05545ec1cc8ab4093247f77275e0743ffed117182eaa9c77877aaac6ac7d35245d1692e8ee1,
}, oid = "1.3.36.3.3.2.8.1.1.9", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP320t1", ShortWeierstrassCurve, {
	"a": 0xd35e472036bc4fb7e13c785ed201e065f98fcfa6f6f40def4f92b9ec7893ec28fcd412b1f1b32e24,
	"b": 0xa7f561e038eb1ed560b3d147db782013064c19f27ed27c6780aaf77fb8a547ceb5b4fef422340353,
	"p": 0xd35e472036bc4fb7e13c785ed201e065f98fcfa6f6f40def4f92b9ec7893ec28fcd412b1f1b32e27,
	"n": 0xd35e472036bc4fb7e13c785ed201e065f98fcfa5b68f12a32d482ec7ee8658e98691555b44c59311,
	"h": 1,
	"Gx": 0x925be9fb01afc6fb4d3e7d4990010f813408ab106c4f09cb7ee07868cc136fff3357f624a21bed52,
	"Gy": 0x63ba3a7a27483ebf6671dbef7abb30ebee084e58a0b077ad42a5a0989d1ee71b1b9bc0455fb0d2c3,
}, oid = "1.3.36.3.3.2.8.1.1.10", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP384r1", ShortWeierstrassCurve, {
	"a": 0x7bc382c63d8c150c3c72080ace05afa0c2bea28e4fb22787139165efba91f90f8aa5814a503ad4eb04a8c7dd22ce2826,
	"b": 0x4a8c7dd22ce28268b39b55416f0447c2fb77de107dcd2a62e880ea53eeb62d57cb4390295dbc9943ab78696fa504c11,
	"p": 0x8cb91e82a3386d280f5d6f7e50e641df152f7109ed5456b412b1da197fb71123acd3a729901d1a71874700133107ec53,
	"n": 0x8cb91e82a3386d280f5d6f7e50e641df152f7109ed5456b31f166e6cac0425a7cf3ab6af6b7fc3103b883202e9046565,
	"h": 1,
	"Gx": 0x1d1c64f068cf45ffa2a63a81b7c13f6b8847a3e77ef14fe3db7fcafe0cbd10e8e826e03436d646aaef87b2e247d4af1e,
	"Gy": 0x8abe1d7520f9c2a45cb1eb8e95cfd55262b70b29feec5864e19c054ff99129280e4646217791811142820341263c5315,
}, oid = "1.3.36.3.3.2.8.1.1.11", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP384t1", ShortWeierstrassCurve, {
	"a": 0x8cb91e82a3386d280f5d6f7e50e641df152f7109ed5456b412b1da197fb71123acd3a729901d1a71874700133107ec50,
	"b": 0x7f519eada7bda81bd826dba647910f8c4b9346ed8ccdc64e4b1abd11756dce1d2074aa263b88805ced70355a33b471ee,
	"p": 0x8cb91e82a3386d280f5d6f7e50e641df152f7109ed5456b412b1da197fb71123acd3a729901d1a71874700133107ec53,
	"n": 0x8cb91e82a3386d280f5d6f7e50e641df152f7109ed5456b31f166e6cac0425a7cf3ab6af6b7fc3103b883202e9046565,
	"h": 1,
	"Gx": 0x18de98b02db9a306f2afcd7235f72a819b80ab12ebd653172476fecd462aabffc4ff191b946a5f54d8d0aa2f418808cc,
	"Gy": 0x25ab056962d30651a114afd2755ad336747f93475b7a1fca3b88f2b6a208ccfe469408584dc2b2912675bf5b9e582928,
}, oid = "1.3.36.3.3.2.8.1.1.12", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP512r1", ShortWeierstrassCurve, {
	"a": 0x7830a3318b603b89e2327145ac234cc594cbdd8d3df91610a83441caea9863bc2ded5d5aa8253aa10a2ef1c98b9ac8b57f1117a72bf2c7b9e7c1ac4d77fc94ca,
	"b": 0x3df91610a83441caea9863bc2ded5d5aa8253aa10a2ef1c98b9ac8b57f1117a72bf2c7b9e7c1ac4d77fc94cadc083e67984050b75ebae5dd2809bd638016f723,
	"p": 0xaadd9db8dbe9c48b3fd4e6ae33c9fc07cb308db3b3c9d20ed6639cca703308717d4d9b009bc66842aecda12ae6a380e62881ff2f2d82c68528aa6056583a48f3,
	"n": 0xaadd9db8dbe9c48b3fd4e6ae33c9fc07cb308db3b3c9d20ed6639cca70330870553e5c414ca92619418661197fac10471db1d381085ddaddb58796829ca90069,
	"h": 1,
	"Gx": 0x81aee4bdd82ed9645a21322e9c4c6a9385ed9f70b5d916c1b43b62eef4d0098eff3b1f78e2d0d48d50d1687b93b97d5f7c6d5047406a5e688b352209bcb9f822,
	"Gy": 0x7dde385d566332ecc0eabfa9cf7822fdf209f70024a57b1aa000c55b881f8111b2dcde494a5f485e5bca4bd88a2763aed1ca2b2fa8f0540678cd1e0f3ad80892,
}, oid = "1.3.36.3.3.2.8.1.1.13", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("brainpoolP512t1", ShortWeierstrassCurve, {
	"a": 0xaadd9db8dbe9c48b3fd4e6ae33c9fc07cb308db3b3c9d20ed6639cca703308717d4d9b009bc66842aecda12ae6a380e62881ff2f2d82c68528aa6056583a48f0,
	"b": 0x7cbbbcf9441cfab76e1890e46884eae321f70c0bcb4981527897504bec3e36a62bcdfa2304976540f6450085f2dae145c22553b465763689180ea2571867423e,
	"p": 0xaadd9db8dbe9c48b3fd4e6ae33c9fc07cb308db3b3c9d20ed6639cca703308717d4d9b009bc66842aecda12ae6a380e62881ff2f2d82c68528aa6056583a48f3,
	"n": 0xaadd9db8dbe9c48b3fd4e6ae33c9fc07cb308db3b3c9d20ed6639cca70330870553e5c414ca92619418661197fac10471db1d381085ddaddb58796829ca90069,
	"h": 1,
	"Gx": 0x640ece5c12788717b9c1ba06cbc2a6feba85842458c56dde9db1758d39c0313d82ba51735cdb3ea499aa77a7d6943a64f7a3f25fe26f06b51baa2696fa9035da,
	"Gy": 0x5b534bd595f5af0fa2c892376c84ace1bb4e3019b71634c01131159cae03cee9d9932184beef216bd71df2dadf86a627306ecff96dbb8bace198b61e00f8b332,
}, oid = "1.3.36.3.3.2.8.1.1.14", origin = "ECC Brainpool"))

cdb.register(_CurveDBEntry("prime192v1", ShortWeierstrassCurve, {
	"a": 0xfffffffffffffffffffffffffffffffefffffffffffffffc,
	"b": 0x64210519e59c80e70fa7e9ab72243049feb8deecc146b9b1,
	"p": 0xfffffffffffffffffffffffffffffffeffffffffffffffff,
	"n": 0xffffffffffffffffffffffff99def836146bc9b1b4d22831,
	"h": 1,
	"Gx": 0x188da80eb03090f67cbf20eb43a18800f4ff0afd82ff1012,
	"Gy": 0x7192b95ffc8da78631011ed6b24cdd573f977a11e794811,
}, aliases = [ "secp192r1", "NIST P-192", "ansip192r1" ], oid = "1.2.840.10045.3.1.1", origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / ANSI X9.62 / FIPS 186-2 / NIST Recommended Elliptic Curves for Federal Government Use"))

cdb.register(_CurveDBEntry("prime192v2", ShortWeierstrassCurve, {
	"a": 0xfffffffffffffffffffffffffffffffefffffffffffffffc,
	"b": 0xcc22d6dfb95c6b25e49c0d6364a4e5980c393aa21668d953,
	"p": 0xfffffffffffffffffffffffffffffffeffffffffffffffff,
	"n": 0xfffffffffffffffffffffffe5fb1a724dc80418648d8dd31,
	"h": 1,
	"Gx": 0xeea2bae7e1497842f2de7769cfe9c989c072ad696f48034a,
	"Gy": 0x6574d11d69b6ec7a672bb82a083df2f2b0847de970b2de15,
}, oid = "1.2.840.10045.3.1.2", origin = "ANSI X9.62"))

cdb.register(_CurveDBEntry("prime192v3", ShortWeierstrassCurve, {
	"a": 0xfffffffffffffffffffffffffffffffefffffffffffffffc,
	"b": 0x22123dc2395a05caa7423daeccc94760a7d462256bd56916,
	"p": 0xfffffffffffffffffffffffffffffffeffffffffffffffff,
	"n": 0xffffffffffffffffffffffff7a62d031c83f4294f640ec13,
	"h": 1,
	"Gx": 0x7d29778100c65a1da1783716588dce2b8b4aee8e228f1896,
	"Gy": 0x38a90f22637337334b49dcb66a6dc8f9978aca7648a943b0,
}, oid = "1.2.840.10045.3.1.3", origin = "ANSI X9.62"))

cdb.register(_CurveDBEntry("prime239v1", ShortWeierstrassCurve, {
	"a": 0x7fffffffffffffffffffffff7fffffffffff8000000000007ffffffffffc,
	"b": 0x6b016c3bdcf18941d0d654921475ca71a9db2fb27d1d37796185c2942c0a,
	"p": 0x7fffffffffffffffffffffff7fffffffffff8000000000007fffffffffff,
	"n": 0x7fffffffffffffffffffffff7fffff9e5e9a9f5d9071fbd1522688909d0b,
	"h": 1,
	"Gx": 0xffa963cdca8816ccc33b8642bedf905c3d358573d3f27fbbd3b3cb9aaaf,
	"Gy": 0x7debe8e4e90a5dae6e4054ca530ba04654b36818ce226b39fccb7b02f1ae,
}, oid = "1.2.840.10045.3.1.4", origin = "ANSI X9.62"))

cdb.register(_CurveDBEntry("prime239v2", ShortWeierstrassCurve, {
	"a": 0x7fffffffffffffffffffffff7fffffffffff8000000000007ffffffffffc,
	"b": 0x617fab6832576cbbfed50d99f0249c3fee58b94ba0038c7ae84c8c832f2c,
	"p": 0x7fffffffffffffffffffffff7fffffffffff8000000000007fffffffffff,
	"n": 0x7fffffffffffffffffffffff800000cfa7e8594377d414c03821bc582063,
	"h": 1,
	"Gx": 0x38af09d98727705120c921bb5e9e26296a3cdcf2f35757a0eafd87b830e7,
	"Gy": 0x5b0125e4dbea0ec7206da0fc01d9b081329fb555de6ef460237dff8be4ba,
}, oid = "1.2.840.10045.3.1.5", origin = "ANSI X9.62"))

cdb.register(_CurveDBEntry("prime239v3", ShortWeierstrassCurve, {
	"a": 0x7fffffffffffffffffffffff7fffffffffff8000000000007ffffffffffc,
	"b": 0x255705fa2a306654b1f4cb03d6a750a30c250102d4988717d9ba15ab6d3e,
	"p": 0x7fffffffffffffffffffffff7fffffffffff8000000000007fffffffffff,
	"n": 0x7fffffffffffffffffffffff7fffff975deb41b3a6057c3c432146526551,
	"h": 1,
	"Gx": 0x6768ae8e18bb92cfcf005c949aa2c6d94853d0e660bbf854b1c9505fe95a,
	"Gy": 0x1607e6898f390c06bc1d552bad226f3b6fcfe48b6e818499af18e3ed6cf3,
}, oid = "1.2.840.10045.3.1.6", origin = "ANSI X9.62"))

cdb.register(_CurveDBEntry("prime256v1", ShortWeierstrassCurve, {
	"a": 0xffffffff00000001000000000000000000000000fffffffffffffffffffffffc,
	"b": 0x5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b,
	"p": 0xffffffff00000001000000000000000000000000ffffffffffffffffffffffff,
	"n": 0xffffffff00000000ffffffffffffffffbce6faada7179e84f3b9cac2fc632551,
	"h": 1,
	"Gx": 0x6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296,
	"Gy": 0x4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5,
}, aliases = [ "secp256r1", "NIST P-256" ], oid = "1.2.840.10045.3.1.7", origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / ANSI X9.62 / FIPS 186-2 / NIST Recommended Elliptic Curves for Federal Government Use"))

cdb.register(_CurveDBEntry("secp112r1", ShortWeierstrassCurve, {
	"a": 0xdb7c2abf62e35e668076bead2088,
	"b": 0x659ef8ba043916eede8911702b22,
	"p": 0xdb7c2abf62e35e668076bead208b,
	"n": 0xdb7c2abf62e35e7628dfac6561c5,
	"h": 1,
	"Gx": 0x9487239995a5ee76b55f9c2f098,
	"Gy": 0xa89ce5af8724c0a23e0e0ff77500,
}, aliases = [ "wap-wsg-idm-ecid-wtls6" ], oid = "1.3.132.0.6", alt_oids = { "wap-wsg-idm-ecid-wtls6": "2.23.43.1.4.6" }, origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / Wireless Application Protocol WAP-261-WTLS-20010406a"))

cdb.register(_CurveDBEntry("secp112r2", ShortWeierstrassCurve, {
	"a": 0x6127c24c05f38a0aaaf65c0ef02c,
	"b": 0x51def1815db5ed74fcc34c85d709,
	"p": 0xdb7c2abf62e35e668076bead208b,
	"n": 0x36df0aafd8b8d7597ca10520d04b,
	"h": 4,
	"Gx": 0x4ba30ab5e892b4e1649dd0928643,
	"Gy": 0xadcd46f5882e3747def36e956e97,
}, oid = "1.3.132.0.7", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp128r1", ShortWeierstrassCurve, {
	"a": 0xfffffffdfffffffffffffffffffffffc,
	"b": 0xe87579c11079f43dd824993c2cee5ed3,
	"p": 0xfffffffdffffffffffffffffffffffff,
	"n": 0xfffffffe0000000075a30d1b9038a115,
	"h": 1,
	"Gx": 0x161ff7528b899b2d0c28607ca52c5b86,
	"Gy": 0xcf5ac8395bafeb13c02da292dded7a83,
}, oid = "1.3.132.0.28", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp128r2", ShortWeierstrassCurve, {
	"a": 0xd6031998d1b3bbfebf59cc9bbff9aee1,
	"b": 0x5eeefca380d02919dc2c6558bb6d8a5d,
	"p": 0xfffffffdffffffffffffffffffffffff,
	"n": 0x3fffffff7fffffffbe0024720613b5a3,
	"h": 4,
	"Gx": 0x7b6aa5d85e572983e6fb32a7cdebc140,
	"Gy": 0x27b6916a894d3aee7106fe805fc34b44,
}, oid = "1.3.132.0.29", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp160k1", ShortWeierstrassCurve, {
	"a": 0,
	"b": 7,
	"p": 0x0fffffffffffffffffffffffffffffffeffffac73,
	"n": 0x100000000000000000001b8fa16dfab9aca16b6b3,
	"h": 1,
	"Gx": 0x03b4c382ce37aa192a4019e763036f4f5dd4d7ebb,
	"Gy": 0x0938cf935318fdced6bc28286531733c3f03c4fee,
}, aliases = [ "ansip160k1" ], oid = "1.3.132.0.9", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp160r1", ShortWeierstrassCurve, {
	"a": 0x0ffffffffffffffffffffffffffffffff7ffffffc,
	"b": 0x01c97befc54bd7a8b65acf89f81d4d4adc565fa45,
	"p": 0x0ffffffffffffffffffffffffffffffff7fffffff,
	"n": 0x100000000000000000001f4c8f927aed3ca752257,
	"h": 1,
	"Gx": 0x04a96b5688ef573284664698968c38bb913cbfc82,
	"Gy": 0x023a628553168947d59dcc912042351377ac5fb32,
}, aliases = [ "ansip160r1" ], oid = "1.3.132.0.8", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp160r2", ShortWeierstrassCurve, {
	"a": 0x0fffffffffffffffffffffffffffffffeffffac70,
	"b": 0x0b4e134d3fb59eb8bab57274904664d5af50388ba,
	"p": 0x0fffffffffffffffffffffffffffffffeffffac73,
	"n": 0x100000000000000000000351ee786a818f3a1a16b,
	"h": 1,
	"Gx": 0x052dcb034293a117e1f4ff11b30f7199d3144ce6d,
	"Gy": 0x0feaffef2e331f296e071fa0df9982cfea7d43f2e,
}, aliases = [ "ansip160r2", "wap-wsg-idm-ecid-wtls7" ], oid = "1.3.132.0.30", alt_oids = { "wap-wsg-idm-ecid-wtls7": "2.23.43.1.4.7" }, origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / Wireless Application Protocol WAP-261-WTLS-20010406a"))

cdb.register(_CurveDBEntry("secp192k1", ShortWeierstrassCurve, {
	"a": 0,
	"b": 3,
	"p": 0xfffffffffffffffffffffffffffffffffffffffeffffee37,
	"n": 0xfffffffffffffffffffffffe26f2fc170f69466a74defd8d,
	"h": 1,
	"Gx": 0xdb4ff10ec057e9ae26b07d0280b7f4341da5d1b1eae06c7d,
	"Gy": 0x9b2f2f6d9c5628a7844163d015be86344082aa88d95e2f9d,
}, aliases = [ "ansip192k1" ], oid = "1.3.132.0.31", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp224k1", ShortWeierstrassCurve, {
	"a": 0,
	"b": 5,
	"p": 0x0fffffffffffffffffffffffffffffffffffffffffffffffeffffe56d,
	"n": 0x10000000000000000000000000001dce8d2ec6184caf0a971769fb1f7,
	"h": 1,
	"Gx": 0x0a1455b334df099df30fc28a169a467e9e47075a90f7e650eb6b7a45c,
	"Gy": 0x07e089fed7fba344282cafbd6f7e319f7c0b0bd59e2ca4bdb556d61a5,
}, aliases = [ "ansip224k1" ], oid = "1.3.132.0.32", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp224r1", ShortWeierstrassCurve, {
	"a": 0xfffffffffffffffffffffffffffffffefffffffffffffffffffffffe,
	"b": 0xb4050a850c04b3abf54132565044b0b7d7bfd8ba270b39432355ffb4,
	"p": 0xffffffffffffffffffffffffffffffff000000000000000000000001,
	"n": 0xffffffffffffffffffffffffffff16a2e0b8f03e13dd29455c5c2a3d,
	"h": 1,
	"Gx": 0xb70e0cbd6bb4bf7f321390b94a03c1d356c21122343280d6115c1d21,
	"Gy": 0xbd376388b5f723fb4c22dfe6cd4375a05a07476444d5819985007e34,
}, aliases = [ "ansip224r1", "NIST P-224", "wap-wsg-idm-ecid-wtls12" ], oid = "1.3.132.0.33", alt_oids = { "wap-wsg-idm-ecid-wtls12": "2.23.43.1.4.12" }, origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / FIPS 186-2 / NIST Recommended Elliptic Curves for Federal Government Use / Wireless Application Protocol WAP-261-WTLS-20010406a"))

cdb.register(_CurveDBEntry("secp256k1", ShortWeierstrassCurve, {
	"a": 0,
	"b": 7,
	"p": 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f,
	"n": 0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141,
	"h": 1,
	"Gx": 0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798,
	"Gy": 0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8,
}, aliases = [ "ansip256k1" ], oid = "1.3.132.0.10", origin = "Certicom Standards for Efficient Cryptography (SEC) 2"))

cdb.register(_CurveDBEntry("secp384r1", ShortWeierstrassCurve, {
	"a": 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000fffffffc,
	"b": 0xb3312fa7e23ee7e4988e056be3f82d19181d9c6efe8141120314088f5013875ac656398d8a2ed19d2a85c8edd3ec2aef,
	"p": 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffff0000000000000000ffffffff,
	"n": 0xffffffffffffffffffffffffffffffffffffffffffffffffc7634d81f4372ddf581a0db248b0a77aecec196accc52973,
	"h": 1,
	"Gx": 0xaa87ca22be8b05378eb1c71ef320ad746e1d3b628ba79b9859f741e082542a385502f25dbf55296c3a545e3872760ab7,
	"Gy": 0x3617de4a96262c6f5d9e98bf9292dc29f8f41dbd289a147ce9da3113b5f0b8c00a60b1ce1d7e819d7a431d7c90ea0e5f,
}, aliases = [ "ansip384r1", "NIST P-384" ], oid = "1.3.132.0.34", origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / FIPS 186-2 / NIST Recommended Elliptic Curves for Federal Government Use"))

cdb.register(_CurveDBEntry("secp521r1", ShortWeierstrassCurve, {
	"a": 0x1fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffc,
	"b": 0x051953eb9618e1c9a1f929a21a0b68540eea2da725b99b315f3b8b489918ef109e156193951ec7e937b1652c0bd3bb1bf073573df883d2c34f1ef451fd46b503f00,
	"p": 0x1ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff,
	"n": 0x1fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffa51868783bf2f966b7fcc0148f709a5d03bb5c9b8899c47aebb6fb71e91386409,
	"h": 1,
	"Gx": 0x0c6858e06b70404e9cd9e3ecb662395b4429c648139053fb521f828af606b4d3dbaa14b5e77efe75928fe1dc127a2ffa8de3348b3c1856a429bf97e7e31c2e5bd66,
	"Gy": 0x11839296a789a3bc0045c8a5fb42c7d1bd998f54449579b446817afbd17273e662c97ee72995ef42640c550b9013fad0761353c7086a272c24088be94769fd16650,
}, aliases = [ "NIST P-521", "ansip521r1" ], oid = "1.3.132.0.35", origin = "Certicom Standards for Efficient Cryptography (SEC) 2 / FIPS 186-2 / NIST Recommended Elliptic Curves for Federal Government Use"))

cdb.register(_CurveDBEntry("wap-wsg-idm-ecid-wtls8", ShortWeierstrassCurve, {
	"a": 0,
	"b": 3,
	"p": 0x0fffffffffffffffffffffffffde7,
	"n": 0x100000000000001ecea551ad837e9,
	"h": 1,
	"Gx": 1,
	"Gy": 2,
}, oid = "2.23.43.1.4.8", origin = "Wireless Application Protocol WAP-261-WTLS-20010406a"))

cdb.register(_CurveDBEntry("wap-wsg-idm-ecid-wtls9", ShortWeierstrassCurve, {
	"a": 0,
	"b": 3,
	"p": 0x0fffffffffffffffffffffffffffffffffffc808f,
	"n": 0x100000000000000000001cdc98ae0e2de574abf33,
	"h": 1,
	"Gx": 1,
	"Gy": 2,
}, oid = "2.23.43.1.4.9", origin = "Wireless Application Protocol WAP-261-WTLS-20010406a"))

cdb.register(_CurveDBEntry("Curve25519", MontgomeryCurve, {
	"a": 486662,
	"b": 1,
	"p": (2 ** 255) - 19,
	"n": (2 ** 252) + 27742317777372353535851937790883648493,
	"h": 8,
	"Gx": 0x9,
	"Gy": 0x5f51e65e475f794b1fe122d388b72eb36dc2b28192839e4dd6163a5d81312c14,
}, origin = "2006 Bernstein"))

# Curve imported from IETF https://tools.ietf.org/html/rfc7748
cdb.register(_CurveDBEntry("Curve448", MontgomeryCurve, {
	"a": 156326,
	"b": 1,
	"p": (2 ** 448) - (2 ** 224) - 1,
	"n": (2 ** 446) - 0x8335dc163bb124b65129c96fde933d8d723a70aadc873d6d54a7bb0d,
	"h": 4,
	"Gx": 0x5,
	"Gy": 0x7d235d1295f5b1f66c98ab6e58326fcecbae5d34f55545d060f75dc28df3f6edb8027e2346430d211312c4b150677af76fd7223d457b5b1a,
}, origin = "2006 Bernstein"))

cdb.register(_CurveDBEntry("Ed25519", TwistedEdwardsCurve, {
	"a": -1,
	"d": 37095705934669439343138083508754565189542113879843219016388785533085940283555,
	"p": (2 ** 255) - 19,
	"n": (2 ** 252) + 27742317777372353535851937790883648493,
	"h": 8,
	"Gx": 0x216936d3cd6e53fec0a4e231fdd6dc5c692cc7609525a7b2c9562d608f25d51a,
	"Gy": 0x6666666666666666666666666666666666666666666666666666666666666658,
}, origin = "2011 Bernstein-Duif-Lange-Schwabe-Yang", quirks = [ CurveQuirkEdDSASetPrivateKeyMSB(), CurveQuirkEdDSAEnsurePrimeOrderSubgroup(), CurveQuirkSigningHashFunction("sha512") ]))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("Anomalous", ShortWeierstrassCurve, {
	"a": 0x98d0fac687d6343eb1a1f595283eb1a1f58d0fac687d635f5e4,
	"b": 0x4a1f58d0fac687d6343eb1a5e2d6343eb1a1f58d0fac688ab3f,
	"p": 0xb0000000000000000000000953000000000000000000001f9d7,
	"n": 0xb0000000000000000000000953000000000000000000001f9d7,
	"h": 1,
	"Gx": 0x101efb35fd1963c4871a2d17edaafa7e249807f58f8705126c6,
	"Gy": 0x22389a3954375834304ba1d509a97de6c07148ea7f5951b20e7,
}, secure = False, origin = "Bernstein http://safecurves.cr.yp.to illustration of additive transfer and small discriminant"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("M-221", MontgomeryCurve, {
	"a": 117050,
	"b": 1,
	"p": 0x1ffffffffffffffffffffffffffffffffffffffffffffffffffffffd,
	"n":  0x40000000000000000000000000015a08ed730e8a2f77f005042605b,
	"h": 8,
	"Gx": 4,
	"Gy":  0xf7acdd2a4939571d1cef14eca37c228e61dbff10707dc6c08c5056d,
}, aliases = [ "Curve2213" ], origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("E-222", TwistedEdwardsCurve, {
	"a": 1,
	"d": 160102,
	"p": 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffff8b,
	"n":  0xffffffffffffffffffffffffffff70cbc95e932f802f31423598cbf,
	"h": 4,
	"Gx": 0x19b12bb156a389e55c9768c303316d07c23adab3736eb2bc3eb54e51,
	"Gy": 28,
}, origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("Curve1174", TwistedEdwardsCurve, {
	"a": 1,
	"d": -1174,
	"p": 0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff7,
	"n": 0x1fffffffffffffffffffffffffffffff77965c4dfd307348944d45fd166c971,
	"h": 4,
	"Gx": 0x37fbb0cea308c479343aee7c029a190c021d96a492ecd6516123f27bce29eda,
	"Gy": 0x6b72f82d47fb7cc6656841169840e0c4fe2dee2af3f976ba4ccb1bf9b46360e,
}, origin = "2013 Bernstein-Hamburg-Krasnova-Lange"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("BN(2,254)", ShortWeierstrassCurve, {
	"a": 0,
	"b": 2,
	"p": 0x2523648240000001ba344d80000000086121000000000013a700000000000013,
	"n": 0x2523648240000001ba344d8000000007ff9f800000000010a10000000000000d,
	"h": 1,
	"Gx": -1,
	"Gy": 1,
}, origin = "2011 Pereira-Simplicio-Naehrig-Barreto"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("ANSSI FRP256v1", ShortWeierstrassCurve, {
	"a": -3,
	"b": 0xee353fca5428a9300d4aba754a44c00fdfec0c9ae4b1a1803075ed967b7bb73f,
	"p": 0xf1fd178c0b3ad58f10126de8ce42435b3961adbcabc8ca6de8fcf353d86e9c03,
	"n": 0xf1fd178c0b3ad58f10126de8ce42435b53dc67e140d2bf941ffdd459c6d655e1,
	"h": 1,
	"Gx": 0xb6b3d4c356c139eb31183d4749d423958c27d2dcaf98b70164c97a2dd98f5cff,
	"Gy": 0x6142e0f7c8b204911f9271f0f3ecef8c2701c307e8e4c9e183115a1554062cfb,
}, oid = "1.2.250.1.223.101.256.1", origin = "Agence nationale de la sécurité des systèmes d'information"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("E-382", TwistedEdwardsCurve, {
	"a": 1,
	"d": -67254,
	"p": 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff97,
	"n":  0xfffffffffffffffffffffffffffffffffffffffffffffffd5fb21f21e95eee17c5e69281b102d2773e27e13fd3c9719,
	"h": 4,
	"Gx": 0x196f8dd0eab20391e5f05be96e8d20ae68f840032b0b64352923bab85364841193517dbce8105398ebc0cc9470f79603,
	"Gy": 17,
}, origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("M-383", MontgomeryCurve, {
	"a": 2065150,
	"b": 1,
	"p": 0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff45,
	"n": 0x10000000000000000000000000000000000000000000000006c79673ac36ba6e7a32576f7b1b249e46bbc225be9071d7,
	"h": 8,
	"Gx": 12,
	"Gy": 0x1ec7ed04aaf834af310e304b2da0f328e7c165f0e8988abd3992861290f617aa1f1b2e7d0b6e332e969991b62555e77e,
}, origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("Curve383187", MontgomeryCurve, {
	"a": 229969,
	"b": 1,
	"p": 0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff45,
	"n": 0x1000000000000000000000000000000000000000000000000e85a85287a1488acd41ae84b2b7030446f72088b00a0e21,
	"h": 8,
	"Gx": 5,
	"Gy": 0x1eebe07dc1871896732b12d5504a32370471965c7a11f2c89865f855ab3cbd7c224e3620c31af3370788457dd5ce46df,
}, origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("Curve41417", TwistedEdwardsCurve, {
	"a": 1,
	"d": 3617,
	"p": 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffef,
	"n":  0x7ffffffffffffffffffffffffffffffffffffffffffffffffffeb3cc92414cf706022b36f1c0338ad63cf181b0e71a5e106af79,
	"h": 8,
	"Gx": 0x1a334905141443300218c0631c326e5fcd46369f44c03ec7f57ff35498a4ab4d6d6ba111301a73faa8537c64c4fd3812f3cbc595,
	"Gy": 34,
}, aliases = [ "Curve3617" ], origin = "2013 Bernstein-Lange"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("Ed448-Goldilocks", TwistedEdwardsCurve, {
	"a": 1,
	"d": -39081,
	"p": 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffffffffffffffffffffffffffffffffffffffffffffffffffff,
	"n": 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffffff7cca23e9c44edb49aed63690216cc2728dc58f552378c292ab5844f3,
	"h": 4,
	"Gx": 0x297ea0ea2692ff1b4faff46098453a6a26adf733245f065c3c59d0709cecfa96147eaaf3932d94c63d96c170033f4ba0c7f0de840aed939f,
	"Gy": 19,
}, origin = "2014 Hamburg", quirks = [ CurveQuirkEdDSASetPrivateKeyMSB(), CurveQuirkEdDSAEnsurePrimeOrderSubgroup(), CurveQuirkSigningHashFunction("shake256-114") ]))

# Curve imported from https://tools.ietf.org/html/rfc8032
cdb.register(_CurveDBEntry("Ed448", TwistedEdwardsCurve, {
	"a": 1,
	"d": -39081,
	"p": 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffeffffffffffffffffffffffffffffffffffffffffffffffffffffffff,
	"n": 0x3fffffffffffffffffffffffffffffffffffffffffffffffffffffff7cca23e9c44edb49aed63690216cc2728dc58f552378c292ab5844f3,
	"h": 4,
	"Gx": 0x4f1970c66bed0ded221d15a622bf36da9e146570470f1767ea6de324a3d3a46412ae1af72ab66511433b80e18b00938e2626a82bc70cc05e,
	"Gy": 0x693f46716eb6bc248876203756c9c7624bea73736ca3984087789c1e05a0c2d73ad3ff1ce67c39c4fdbd132c4ed7c8ad9808795bf230fa14,
}, origin = "https://tools.ietf.org/html/rfc8032", quirks = [ CurveQuirkEdDSASetPrivateKeyMSB(), CurveQuirkEdDSAEnsurePrimeOrderSubgroup(), CurveQuirkSigningHashFunction("shake256-114") ]))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("M-511", MontgomeryCurve, {
	"a": 530438,
	"b": 1,
	"p": 0x7fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff45,
	"n": 0x100000000000000000000000000000000000000000000000000000000000000017b5feff30c7f5677ab2aeebd13779a2ac125042a6aa10bfa54c15bab76baf1b,
	"h": 8,
	"Gx": 5,
	"Gy": 0x2fbdc0ad8530803d28fdbad354bb488d32399ac1cf8f6e01ee3f96389b90c809422b9429e8a43dbf49308ac4455940abe9f1dbca542093a895e30a64af056fa5,
}, aliases = [ "Curve511187" ], origin = "2013 Aranha-Barreto-Pereira-Ricardini"))

# Curve imported from SafeCurves http://safecurves.cr.yp.to
cdb.register(_CurveDBEntry("E-521", TwistedEdwardsCurve, {
	"a": 1,
	"d": -376014,
	"p": 0x1ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff,
	"n":  0x7ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffd15b6c64746fc85f736b8af5e7ec53f04fbd8c4569a8f1f4540ea2435f5180d6b,
	"h": 4,
	"Gx":  0x752cb45c48648b189df90cb2296b2878a3bfd9f42fc6c818ec8bf3c9c0c6203913f6ecc5ccc72434b1ae949d568fc99c6059d0fb13364838aa302a940a2f19ba6c,
	"Gy": 12,
}, origin = "2013 Bernstein-Lange / 2013 Hamburg / 2013 Aranha-Barreto-Pereira-Ricardini"))

cdb.register(_CurveDBEntry("rigol", ShortWeierstrassCurve, {
	"a": 0x2982,
	"b": 0x3408,
	"p": 0xaebf94cee3e707,
	"n": 0xaebf94d5c6aa71,
	"h": 1,
	"Gx": 0x7a3e808599a525,
	"Gy": 0x28be7fafd2a052,
}, origin = "Rigol DS2xxx feature activation curve"))

def getcurvedb():
	"""Returns an instance of the curve database singleton object."""
	return CurveDB()

def getcurvenames():
	"""Returns the names of all curves known to toyecc."""
	return CurveDB().curvenames()

def getcurveentry(name):
	"""Returns a curve entry by its name."""
	return CurveDB().getentry(name)

def getcurvebyname(name):
	"""Returns a curve by its name."""
	return CurveDB()[name]



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

from .Polynomial import Polynomial

class DivisionPolynomial(object):
	def __init__(self, curve):
		"""Creates a division polynomial generator which returns \psi_i for the
		given curve in Weierstrass form."""
		self._curve = curve
		assert(self._curve.curvetype == "shortweierstrass")
		self._cache = { }
		self._curvepoly = None
		self._initcache()

	def _initcache(self):
		(a, b) = (self.curve.a, self.curve.b)
		x = Polynomial(self.curve.p)
		self._cache[0] = Polynomial(self.curve.p, 0)
		self._cache[1] = Polynomial(self.curve.p, 1)
		self._cache[2] = Polynomial(self.curve.p, 2)
		self._cache[3] = (3 * x**4) + (6 * a * x**2) + (12 * b * x) - (a**2)
		self._cache[4] = 4 * (x**6 + (5 * a * x**4) + (20 * b * x**3) - (5 * a**2 * x**2) - (4 * a * b * x) - (8 * b**2) - (a**3))
		self._curvepoly = x**3 + (a * x) + b

	@property
	def curve(self):
		return self._curve

	def __getitem__(self, index):
		if index not in self._cache:
			m = index // 2
			if (index % 2) == 1:
				# The paper says this would be correct:
				# result = (self[m + 2] * self[m]**3) - (self[m - 1] * self[m + 1] ** 3)
				# But MIRACL does it differently. Use the MIRACL approach:
				if (m % 2) == 0:
					result = (self._curvepoly**2 * self[m + 2] * self[m]**3) - (self[m - 1] * self[m + 1]**3)
				else:
					result = (self[m + 2] * self[m]**3) - (self._curvepoly**2 * self[m - 1] * self[m + 1]**3)
			else:
				result = (self[m] // 2) * ((self[m + 2] * self[m - 1]**2) - (self[m - 2] * self[m + 1]**2))
			self._cache[index] = result
		return self._cache[index]

	def __str__(self):
		return "DivisionPolys<%s, %d cached>" % (str(self.curve), len(self._cache))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2016 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

from .FieldElementTests import FieldElementTests
from .FieldElementSqrtTests import FieldElementSqrtTests
from .ECTests import ECTests
from .CryptoOpsTests import CryptoOpsTests
from .CurveTests import CurveTests
from .Ed25519BasicTests import Ed25519BasicTests
from .Ed25519ExtdTests import Ed25519ExtdTests
from .TwEdMontConversionTests import TwEdMontConversionTests
from .TwEdMontDomainTests import TwEdMontDomainTests

Ed25519ExtdTests.set_test_scope("full")



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2016 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

from .PubKeyOps import PubKeyOpECDSAVerify, PubKeyOpECDSAExploitReusedNonce, PubKeyOpEDDSAVerify, PubKeyOpEDDSAEncode, PubKeyOpECIESEncrypt, PubKeyOpLoad

class ECPublicKey(PubKeyOpECDSAVerify, PubKeyOpECDSAExploitReusedNonce, PubKeyOpEDDSAVerify, PubKeyOpEDDSAEncode, PubKeyOpECIESEncrypt, PubKeyOpLoad):
	"""Elliptic curve public key abstraction. An EC public key is just a point
	on the curve, which is why the constructor only takes this (public) point
	as a parameter. The public key abstraction allows this point to be used in
	various meaningful purposes (ECDSA signature verification, etc.)."""

	def __init__(self, point):
		self._point = point

	@property
	def curve(self):
		return self._point.curve

	@property
	def point(self):
		return self._point

	def __str__(self):
		return "PublicKey<%s>" % (str(self.point))



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2016 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
import random
from .. import getcurvebyname

class TwEdMontConversionTests(unittest.TestCase):
	def setUp(self):
		self._mont = getcurvebyname("curve25519")
		self._twed = getcurvebyname("ed25519")

	def test_generator(self):
		Ge = self._twed.G
		self.assertEqual(Ge.convert(self._mont), self._mont.G)

		Gm = self._mont.G
		self.assertEqual(Gm.convert(self._twed), self._twed.G)

	def test_neutral(self):
		Oe = self._twed.neutral()
		self.assertEqual(Oe.convert(self._mont), self._mont.neutral())

		Om = self._mont.neutral()
		self.assertEqual(Om.convert(self._twed), self._twed.neutral())

	def test_scalar_conv_twed_to_mont(self):
		for i in range(2, 1000, 17):
			Pe = self._twed.G * i
			Pm = self._mont.G * i
			self.assertEqual(Pe.convert(self._mont), Pm)

	def test_scalar_conv_mont_to_twed(self):
		for i in range(2, 1000, 17):
			Pe = self._twed.G * i
			Pm = self._mont.G * i
			self.assertEqual(Pm.convert(self._twed), Pe)

	def test_random_ptadds(self):
		P = self._twed.G
		scalar = 1
		self.assertEqual(P, self._twed.G * scalar)
		for i in range(20):
			P = P.convert(self._mont)
			self.assertEqual(P, self._mont.G * scalar)

			r = random.randint(2 ** 32, 2 ** 64)
			Q = self._mont.G * r
			P = P + Q
			scalar += r
			self.assertEqual(P, self._mont.G * scalar)

			P = P.convert(self._twed)
			self.assertEqual(P, self._twed.G * scalar)

			r = random.randint(2 ** 32, 2 ** 64)
			Q = self._twed.G * r
			P = P + Q
			scalar += r
			self.assertEqual(P, self._twed.G * scalar)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
import random
from ..FieldElement import FieldElement

class FieldElementSqrtTests(unittest.TestCase):
	_PRIME_1_MOD_4 = [
		10153776240248910961,
		13313520292754238121,
		15378050022937467689,
		15438372505554348001,
		15476752070461085857,
		17875746998751974477,
		18057875082506157121,
	]

	_PRIME_3_MOD_4 = [
		10164036603611688719,
		11802848526449265743,
		12398748969631217971,
		12821403563901948251,
		12871176861409354567,
		13022588269702988959,
		13784119435837260871,
	]

	def test_integrity(self):
		for p in self._PRIME_3_MOD_4:
			self.assertEqual(p % 4, 3)

	def _test_primes(self, primeset, testcnt = 30):
		for p in primeset:
			rootable = False
			for i in range(testcnt):
				qsqr = FieldElement(random.randrange(p), p)
				q = qsqr.sqrt()
				if q is not None:
					rootable = True
					(a, b) = q
					self.assertEqual(a * a, qsqr)
					self.assertEqual(b * b, qsqr)
			self.assertTrue(rootable)

	def test_basic(self):
		self._test_primes(self._PRIME_3_MOD_4)

	def test_extd(self):
		self._test_primes(self._PRIME_1_MOD_4)

	def test_small_body(self):
		p = 263
		for i in range(1, p):
			i = FieldElement(i, p)
			q = i.sqr()
			r = q.sqrt()
			self.assertEqual(r[0] * r[0], q)
			self.assertEqual(r[1] * r[1], q)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import hashlib
import collections

from .FieldElement import FieldElement
from .Random import secure_rand, secure_rand_int_between
from .AffineCurvePoint import AffineCurvePoint
from .CurveDB import CurveDB
from .ShortWeierstrassCurve import ShortWeierstrassCurve
from .ASN1 import parse_asn1_private_key, parse_asn1_field_params_fp
from . import Tools
from .CurveQuirks import CurveQuirkEdDSASetPrivateKeyMSB, CurveQuirkEdDSAEnsurePrimeOrderSubgroup, CurveQuirkSigningHashFunction

class PrivKeyOpECDSASign(object):
	ECDSASignature = collections.namedtuple("ECDSASignature", [ "hashalg", "r", "s" ])

	def ecdsa_sign_hash(self, message_digest, k = None, digestname = None):
		"""Signs a given messagedigest, given as bytes, using ECDSA.
		Optionally a nonce k can be supplied which should usually be unqiuely
		chosen for every ECDSA signature. This way it is possible to
		deliberately create broken signatures which can be exploited later on.
		If k is not supplied, it is randomly chosen. If a digestname is
		supplied the name of this digest eventually ends up in the
		ECDSASignature object."""
		assert(isinstance(message_digest, bytes))
		assert((k is None) or isinstance(k, int))

		# Convert message digest to integer value
		e = Tools.ecdsa_msgdigest_to_int(message_digest, self.curve.n)

		# Select a random integer (if None is supplied!)
		if k is None:
			k = secure_rand_int_between(1, self.curve.n - 1)

		# r = (k * G)_x mod n
		Rmodp = k * self.curve.G
		r = int(Rmodp.x) % self.curve.n
		assert(r != 0)

		s = FieldElement(e + self.scalar * r, self.curve.n) // k

		return self.ECDSASignature(r = r, s = int(s), hashalg = digestname)

	def ecdsa_sign(self, message, digestname, k = None):
		"""Signs a given message with the digest that is given as a string.
		Optionally a nonce k can be supplied which should usually be unqiuely
		chosen for every ECDSA signature. This way it is possible to
		deliberately create broken signatures which can be exploited later
		on. If k is not supplied, it is randomly chosen."""
		assert(isinstance(message, bytes))
		assert(isinstance(digestname, str))
		digest_fnc = hashlib.new(digestname)
		digest_fnc.update(message)
		message_digest = digest_fnc.digest()
		return self.ecdsa_sign_hash(message_digest, k = k, digestname = digestname)


class PrivKeyOpECIESDecrypt(object):
	def ecies_decrypt(self, R):
		"""Takes the transmitted point R and reconstructs the shared secret
		point S using the private key."""
		# Transmitted R is given, restore the symmetric key S
		return self._scalar * R


class PrivKeyOpEDDSASign(object):
	class EDDSASignature(object):
		def __init__(self, curve, R, s):
			self._curve = curve
			self._R = R
			self._s = s

		@property
		def curve(self):
			return self._curve

		@property
		def R(self):
			return self._R

		@property
		def s(self):
			return self._s

		def encode(self):
			"""Performs serialization of the signature as used by EdDSA."""
			return self.R.eddsa_encode() + Tools.inttobytes_le(self.s, (self.curve.B + 7) // 8)

		@classmethod
		def decode(cls, curve, encoded_signature):
			"""Performs deserialization of the signature as used by EdDSA."""
			assert(isinstance(encoded_signature, bytes))
			coordlen = (curve.B + 7) // 8
			assert(len(encoded_signature) == 2 * coordlen)
			encoded_R = encoded_signature[:coordlen]
			encoded_s = encoded_signature[coordlen:]
			R = AffineCurvePoint.eddsa_decode(curve, encoded_R)
			s = Tools.bytestoint_le(encoded_s)
			return cls(curve, R, s)

		def __eq__(self, other):
			return (self.R, self.s) == (other.R, other.s)

		def __str__(self):
			return "EDDSASignature<R = %s, s = %s>" % (self.R, self.s)

	def eddsa_sign(self, message):
		"""Performs an EdDSA signature of the message. For this to work the
		curve has to be a twisted Edwards curve and the private key scalar has
		to be generated from a hashed seed. This hashed seed is automatically
		generated when a keypair is generated using, for example, the
		eddsa_generate() function instead of the regular key generation
		function generate()."""
		assert(self.curve.curvetype == "twistededwards")
		if self._seed is None:
			raise Exception("EdDSA requires a seed which is the source for calculation of the private key scalar.")
		if not self.curve.has_quirk(CurveQuirkSigningHashFunction):
			raise Exception("Unable to determine EdDSA signature function.")

		quirk = self.curve.get_quirk(CurveQuirkSigningHashFunction)
		h = quirk.hashdata(self._seed)

		coordlen = (self.curve.B + 7) // 8
		r = Tools.bytestoint_le(quirk.hashdata(h[coordlen : 2 * coordlen] + message))
		R = r * self.curve.G
		s = (r + Tools.bytestoint_le(quirk.hashdata(R.eddsa_encode() + self.pubkey.point.eddsa_encode() + message)) * self.scalar) % self.curve.n
		sig = self.EDDSASignature(self.curve, R, s)
		return sig


class PrivKeyOpEDDSAKeyGen(object):
	@classmethod
	def eddsa_generate(cls, curve, seed = None):
		"""Generates a randomly selected seed value. This seed value is then
		hashed using the EdDSA hash function (SHA512 for Ed2556 and
		Shake256-114 for Ed448) and the resulting value is (slightly modified)
		used as the private key scalar.  Since for EdDSA signing operations
		this seed value is needed, it is also stored within the private key."""
		coordlen = (curve.B + 7) // 8
		if seed is None:
			seed = secure_rand(coordlen)
		assert(isinstance(seed, bytes))
		assert(len(seed) == coordlen)

		# Calculate hash over seed and generate scalar from hash over seed
		if not curve.has_quirk(CurveQuirkSigningHashFunction):
			raise Exception("Unable to determine EdDSA signature function.")
		quirk = curve.get_quirk(CurveQuirkSigningHashFunction)
		h = quirk.hashdata(seed)
		a = int.from_bytes(h[:coordlen], byteorder = "little") & ((1 << (curve.B - 1)) - 1)

		# Do we need to mask out lower significant bits to ensure that we use a
		# prime order subgroup?
		if curve.has_quirk(CurveQuirkEdDSAEnsurePrimeOrderSubgroup):
			if not Tools.is_power_of_two(curve.h):
				raise Exception("Can only ensure prime order subgroup by masking 'a' when curve cofactor is a power of two, h = %d isn't." % (curve.h))
			a &= ~(curve.h - 1)

		# Is the MSB of the curve always set to ensure constant runtime of the
		# Montgomery ladder?
		if curve.has_quirk(CurveQuirkEdDSASetPrivateKeyMSB):
			bit = curve.n.bit_length() + 1
			a |= (1 << bit)
		privkey = cls(a, curve)
		privkey.set_seed(seed)
		return privkey


class PrivKeyOpEDDSAEncode(object):
	def eddsa_encode(self):
		"""Performs serialization of a private key that is used for EdDSA."""
		return self.seed

	@classmethod
	def eddsa_decode(cls, curve, encoded_privkey):
		"""Performs decoding of a serialized private key as it is used for EdDSA."""
		return cls.eddsa_generate(curve, encoded_privkey)


class PrivKeyOpECDH(object):
	def ecdh_compute(self, peer_pubkey):
		"""Compute the shared secret point using our own private key and the
		public key of our peer."""
		return self.scalar * peer_pubkey.point


class PrivKeyOpLoad(object):
	@classmethod
	def load_derdata(cls, derdata):
		"""Loads an EC private key from a DER-encoded ASN.1 bytes object."""
		asn1 = parse_asn1_private_key(derdata)
		private_key_scalar = Tools.bytestoint(asn1["privateKey"])
		curve = CurveDB().get_curve_from_asn1(asn1["parameters"])
		return cls(private_key_scalar, curve)

	@classmethod
	def load_pem(cls, pemfilename):
		"""Loads an EC private key from a PEM-encoded 'EC PRIVATE KEY' file."""
		return cls.load_derdata(Tools.load_pem_data(pemfilename, "EC PRIVATE KEY"))

	@classmethod
	def load_der(cls, derfilename):
		"""Loads an EC private key from a DER-encoded ASN.1 file."""
		with open(derfilename, "rb") as f:
			data = f.read()
			return cls.load_derdata(data)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
from .. import getcurvebyname, ShortWeierstrassCurve, AffineCurvePoint

class TwistTests(unittest.TestCase):
	def test_brainpool_gf_p_isomorphism(self):
		curve_lengths = (160, 192, 224, 256, 320, 384, 512)
		for bitlen in curve_lengths:
			name_orig = "brainpoolP%dr1" % (bitlen)
			name_twist = "brainpoolP%dt1" % (bitlen)
			curve = getcurvebyname(name_orig)
			twist = getcurvebyname(name_twist)
			self.assertTrue(curve.is_isomorphous_curve(twist))

			my_twist = curve.twist_fp_isomorphic_fixed_a(-3)
			self.assertEqual(my_twist.a, -3)
			self.assertEqual(my_twist.b, twist.b)
			self.assertTrue(twist.is_isomorphous_curve(my_twist))

	def test_brainpool_twist(self):
		curve = getcurvebyname("secp112r1")

		# Known twists as calculated by SAGE
		known_twist_1 = ShortWeierstrassCurve.init_rawcurve(0xdb7c2abf62e35e668076bead1bdb, 0xd6ce5ea491322bfe05c0d7022be3, curve.p)
		known_twist_2 = ShortWeierstrassCurve.init_rawcurve(0xdb7c2abf62e35e668076bead19cb, 0x684a8c3bd69e6065510ec0eb9900, curve.p)
		known_twist_3 = ShortWeierstrassCurve.init_rawcurve(0xdb7c2abf62e35e668076bead148b, 0x6d03b63b43341ad58bb247c1729f, curve.p)
		for i in range(10):
			twist = curve.twist()
			self.assertTrue(twist.is_isomorphous_curve(known_twist_1))
			self.assertTrue(twist.is_isomorphous_curve(known_twist_2))
			self.assertTrue(twist.is_isomorphous_curve(known_twist_3))

	def test_twist_n_h(self):
		# Cofactor taken from https://safecurves.cr.yp.to/twist.html
		curve = getcurvebyname("brainpoolP256t1")
		twist = curve.twist()

		twist_curve_order = curve.p + 1 + curve.frobenius_trace

		h = 5**2 * 175939 * 492167257 * 8062915307 * 2590895598527 * 4233394996199
		n = twist_curve_order // h
		assert((twist_curve_order % h) == 0)

#		TODO: Not sure what the point of this test is now that we don't have a
#		twisted generator point G
#		full_twist = ShortWeierstrassCurve(a = int(twist.a), b = int(twist.b), p = twist.p, n = n, h = h, Gx = int(twist.G.x), Gy = int(twist.G.y))

	def test_twist_add(self):
		curve = getcurvebyname("secp112r1").twist(32)
		P1 = AffineCurvePoint(0x6229b962d45ca6852e375adc80e1, 0x1633adc1c8f0e611552278a24ada, curve)
		P2 = AffineCurvePoint(0x89f14abd7ce333e5abf5fd25293f, 0x5637cb20fe3c2d12b965fd03e1a6, curve)
		P3 = AffineCurvePoint(0xbd2fab84d3fef8f637ef1c341e00, 0x3bd95f46f30cccb31f4e4b6d6595, curve)

		self.assertTrue(P1.oncurve())
		self.assertTrue(P2.oncurve())
		self.assertTrue(P3.oncurve())
		self.assertEqual(P1 + P1, P2)
		self.assertEqual(P1 + P2, P3)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
from .. import getcurvebyname, getcurvenames
from ..ECPrivateKey import ECPrivateKey

class CryptoOpsTests(unittest.TestCase):
	def test_curve_integrity(self):
		for curvename in getcurvenames():
			curve = getcurvebyname(curvename)
			self.assertTrue(curve.G.oncurve())

	def test_simple_sign_verify(self):
		curve = getcurvebyname("secp112r1")
		privkey = ECPrivateKey(0xdeadbeef, curve)
		self.assertEqual(int(privkey.pubkey.point.x), 3029259716094196738484362740763961)
		self.assertEqual(int(privkey.pubkey.point.y), 2918181739692718713384134377830669)

		msg = b"foobar"
		signature = privkey.ecdsa_sign(msg, "sha1", k = 12345)
		self.assertEqual(signature.s, 1960761230049936699759766101723490)
		self.assertEqual(signature.r, 1696427335541514286367855701829018)

		self.assertTrue(privkey.pubkey.ecdsa_verify(msg, signature))

	def test_openssl_signature(self):
		curve = getcurvebyname("prime239v1")
		msg = b"foobar"
		r = 0x76EAA8341A6E30FE72C87343A50DD2BBE472CF5E5A394D245DF354EF11CD
		s = 0x0304117012F37FC316D9030E456F53940008C7FAF569BBE6E5B4AA1A2D40
		hashalg = "sha256"

		privkeystr = "6e:29:dc:d3:c4:7c:a3:3c:3d:0e:f5:3a:15:14:18:ff:ff:9b:44:dc:26:21:ac:d3:fc:ac:ca:5b:c8:f2"
		privkeyint = int(privkeystr.replace(":", ""), 16)
		privkey = ECPrivateKey(privkeyint, curve)
		self.assertEqual(privkey.pubkey.point.x, 0x0f5e2baa05719806cbdb133d438011efc2c4e036d59cb799fa87d2ecab97)
		self.assertEqual(privkey.pubkey.point.y, 0x56fd18e031e2869a872bbc96ec473bd5eb2e9a22a2e246bd00ed0491fdde)

		signature = ECPrivateKey.ECDSASignature(r = r, s = s, hashalg = hashalg)
		self.assertTrue(privkey.pubkey.ecdsa_verify(msg, signature))

	def test_ecies(self):
		curve = getcurvebyname("secp112r1")
		privkey = ECPrivateKey.generate(curve)

		encparams = privkey.pubkey.ecies_encrypt()
		S = privkey.ecies_decrypt(encparams["R"])
		self.assertEqual(S, encparams["S"])


	def test_ecdh(self):
		curve = getcurvebyname("secp112r1")

		party1_privkey = ECPrivateKey.generate(curve)
		party2_privkey = ECPrivateKey.generate(curve)

		S1 = party1_privkey.ecdh_compute(party2_privkey.pubkey)
		S2 = party2_privkey.ecdh_compute(party1_privkey.pubkey)
		self.assertEqual(S1, S2)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2016 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
from .. import FieldElement, getcurvebyname, ECPublicKey, ECPrivateKey

class Ed25519BasicTests(unittest.TestCase):
	def test_sign_verify(self):
		curve = getcurvebyname("ed25519")
		privkey = ECPrivateKey.eddsa_generate(curve)

		msg = b"foobar"
		signature = privkey.eddsa_sign(msg)

		self.assertTrue(privkey.pubkey.eddsa_verify(msg, signature))
		self.assertFalse(privkey.pubkey.eddsa_verify(msg + b"x", signature))

	def test_sig_encode_decode(self):
		curve = getcurvebyname("ed25519")
		privkey = ECPrivateKey.eddsa_generate(curve)
		msg = b"foobar"
		signature = privkey.eddsa_sign(msg)

		encoded_signature = signature.encode()
		decoded_signature = ECPrivateKey.EDDSASignature.decode(curve, encoded_signature)
		self.assertEqual(decoded_signature, signature)
		self.assertTrue(privkey.pubkey.eddsa_verify(msg, signature))
		self.assertTrue(privkey.pubkey.eddsa_verify(msg, decoded_signature))

	def test_seeding_signing(self):
		curve = getcurvebyname("ed25519")
		seed = bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e")

		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 55506792121812326601863127824214732996787216741471646079326674346494898512112)

		pubkey = privkey.pubkey
		self.assertEqual(pubkey.point.x, 0x7de7bc3e0b0c077c2a104623a8e66cb7d9ffb3b1a594969bb55997dbe66d4264)
		self.assertEqual(pubkey.point.y, 0x704f7339f697f88d190052a76091805f24c77aa232acf69b19082fa036967744)

		msg = b"Foobar!"
		signature = privkey.eddsa_sign(msg)

		self.assertEqual(signature.R.x, 0x7fc3a9aca15bb635bc471c6fd410f12e3658e0b226452f025992c931ab61fc18)
		self.assertEqual(signature.R.y, 0x3a5902651bf0eb71dfbb604d3e20511821f7e720b5814e8a9b744aede20e7986)
		self.assertEqual(signature.s, 1517157819819635474038904401642496669072502937558957834094270557835510822428)


	def test_key_encoding(self):
		curve = getcurvebyname("ed25519")
		seed = bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e")

		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)

		self.assertEqual(privkey.eddsa_encode(), bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e"))
		self.assertEqual(privkey.pubkey.eddsa_encode(), bytes.fromhex("44779636a02f08199bf6ac32a27ac7245f809160a75200198df897f639734f70"))

	def test_key_decoding(self):
		curve = getcurvebyname("ed25519")
		seed = bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e")

		privkey = ECPrivateKey.eddsa_decode(curve, bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e"))
		self.assertEqual(privkey.seed, seed)
		self.assertEqual(privkey.scalar, 55506792121812326601863127824214732996787216741471646079326674346494898512112)

		pubkey = ECPublicKey.eddsa_decode(curve, bytes.fromhex("44779636a02f08199bf6ac32a27ac7245f809160a75200198df897f639734f70"))
		self.assertEqual(pubkey.point, privkey.pubkey.point)


	def test_sig_encoding(self):
		curve = getcurvebyname("ed25519")
		seed = bytes.fromhex("5da0ed08799092411e90140e7b86058276fe293efd40afa816bc0ccc3f43492e")

		privkey = ECPrivateKey.eddsa_generate(curve, seed = seed)

		msg = b"Foobar!"
		signature = privkey.eddsa_sign(msg)

		self.assertEqual(signature.encode(), bytes.fromhex("86790ee2ed4a749b8a4e81b520e7f7211851203e4d60bbdf71ebf01b6502593a1c36ca859621a386dfaa79c5a24bb6eef1111c76bfb7fc0c48491b2b4fae5a03"))

	def test_sig_decoding(self):
		curve = getcurvebyname("ed25519")
		encoded_sig = bytes.fromhex("86790ee2ed4a749b8a4e81b520e7f7211851203e4d60bbdf71ebf01b6502593a1c36ca859621a386dfaa79c5a24bb6eef1111c76bfb7fc0c48491b2b4fae5a03")

		signature = ECPrivateKey.EDDSASignature.decode(curve, encoded_sig)
		self.assertEqual(signature.R.x, 0x7fc3a9aca15bb635bc471c6fd410f12e3658e0b226452f025992c931ab61fc18)
		self.assertEqual(signature.R.y, 0x3a5902651bf0eb71dfbb604d3e20511821f7e720b5814e8a9b744aede20e7986)
		self.assertEqual(signature.s, 1517157819819635474038904401642496669072502937558957834094270557835510822428)



#
#	toyecc - A small Elliptic Curve Cryptography Demonstration.
#	Copyright (C) 2011-2022 Johannes Bauer
#
#	This file is part of toyecc.
#
#	toyecc is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	toyecc is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with toyecc; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#

import unittest
import random
from .. import getcurvebyname
from ..AffineCurvePoint import AffineCurvePoint

class PointSerializationTests(unittest.TestCase):
	def _test_curve_point(self, point):
		Pser = point.serialize_uncompressed()
		Pdeser = AffineCurvePoint.deserialize_uncompressed(Pser, point.curve)
		self.assertEqual(Pdeser, point)

	def _test_curve_points(self, curve, expect_points):
		for (scalar, expect_encoding) in expect_points.items():
			P = scalar * curve.G
			Pser = P.serialize_uncompressed()
			self.assertEqual(Pser, expect_encoding)
			deserialized = AffineCurvePoint.deserialize_uncompressed(Pser, curve)
			self.assertEqual(deserialized, P)

	def _get_low_point(self, curve):
		for i in range(1, 1000):
			P = curve.getpointwithx(i)
			if P is not None:
				return P[0]

	def _get_random_point(self, curve):
		for i in range(1, 1000):
			P = curve.getpointwithx(random.randint(1, curve.p))
			if P is not None:
				return P[0]

	def test_secp112r1(self):
		expect_points = {
			0x38b320d1a8c75a275a8c6c6d8df3: bytes.fromhex("040020b65a4224192dab5255eee17d470a3471917b1672ed341a571b73"),
			0x65cd8dbdbd38d7a98a064e4cf741: bytes.fromhex("04000024021cce683926bf0c6ac6b0cd71d968563e2646d9382e27b801"),
		}
		curve = getcurvebyname("secp112r1")
		self._test_curve_points(curve, expect_points)

	def test_secp160r1(self):
		# secp160r1 has 160 bit (20 byte) p, but 161 bit (21 bytes) n
		expect_points = {
			0x415bb31e0096c6255f26618752583f6cb43b5950: bytes.fromhex("0400b80c796da4b9712f52b0a4c18b0dd269d00b473259900e73bc0b8b281dd36dfa0ad36c36f04a99"),
			0x29d5dfd3c7dac354209648ed852bc09f99fdc5e1: bytes.fromhex("0400655731f18c3345cc57e5d1a0d2109e82622896d3ea2928bfab8b74b060a4984d25413a37d232f1"),
			0x1c7929dab7fb01329952d4705608312505d3a5f8: bytes.fromhex("04007ade8f2dbc473b82ff37bab7443eaf7be5e15e2cf24fef5cc3f0e5f3533327c99814c595c44080"),
		}
		curve = getcurvebyname("secp160r1")
		self._test_curve_points(curve, expect_points)

	def test_random_serialization_deserialization(self):
		curve = getcurvebyname("secp112r1")
		for i in range(100):
			P = self._get_random_point(curve)
			self.assertTrue(P.oncurve())
			self._test_curve_point(P)

	def test_low_point(self):
		for curvename in [ "secp112r1", "secp160r1" ]:
			curve = getcurvebyname(curvename)
			self._test_curve_point(self._get_low_point(curve))
